import time
from typing import Optional

import numpy as np

from .base_callback import BaseCallback


class ConvergenceTimeGetter(BaseCallback):
	"""
	Monitor the training process and return the time it took to pass the threshold.
	"""
	def __init__(
			self,
			*,
			metric: str,
			threshold: float,
			minimize_metric: bool,
			**kwargs
	):
		"""
		Constructor for ConvergenceTimeGetter class.
		
		:param metric: Name of the metric to monitor.
		:type metric: str
		:param threshold: Threshold value for the metric.
		:type threshold: float
		:param minimize_metric: Whether to minimize or maximize the metric.
		:type minimize_metric: bool
		:param kwargs: The keyword arguments to pass to the BaseCallback.
		"""
		super().__init__(**kwargs)
		self.threshold = threshold
		self.metric = metric
		self.minimize_metric = minimize_metric
		self.threshold_met = False
		self.time_convergence = np.inf
		self.itr_convergence = np.inf
		self.training_time = np.inf
		self.start_time = None
	
	def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
		if self.save_state:
			state = checkpoint.get(self.name, {})
			if state.get("threshold") == self.threshold and state.get("metric") == self.metric:
				super().load_checkpoint_state(trainer, checkpoint)
				self.start_time = time.time()
				if np.isfinite(state.get("time_convergence")):
					self.start_time -= state.get("training_time", 0)
			# TODO: change start time and add training time, etc.
	
	def start(self, trainer, **kwargs):
		super().start(trainer)
		self.start_time = time.time()
	
	def close(self, trainer, **kwargs):
		self.training_time = time.time() - self.start_time
	
	def on_iteration_end(self, trainer, **kwargs):
		if not self.threshold_met:
			if self.minimize_metric:
				self.threshold_met = trainer.current_training_state.itr_metrics[self.metric] < self.threshold
			else:
				self.threshold_met = trainer.current_training_state.itr_metrics[self.metric] > self.threshold
			if self.threshold_met:
				self.time_convergence = time.time() - self.start_time
				self.itr_convergence = trainer.current_training_state.iteration
	
	def __repr__(self):
		repr_str = f"ConvergenceTimeGetter("
		repr_str += f"metric={self.metric}, "
		repr_str += f"threshold={self.threshold}, "
		repr_str += f"minimize_metric={self.minimize_metric})"
		repr_str += f"<time_convergence={self.time_convergence} [s], "
		repr_str += f"itr_convergence={self.itr_convergence}, "
		repr_str += f"training_time={self.training_time} [s]>"
		return repr_str






import os.path
import pickle
import time
import multiprocessing as mp
from typing import Union, Optional

from matplotlib import pyplot as plt

from .base_callback import BaseCallback


class _VisualizerProcess(mp.Process):
	"""
	Process to load a training history from pickle file and plot it. The process will update the plot every `update_dt`
	seconds and will close when join method is called.
	
	:Attributes:
		- **fig** (plt.Figure): The figure to plot.
		- **axes** (plt.Axes): The axes to plot.
		- **lines** (list): The list of lines to plot.
	"""
	def __init__(self, history_path: str, lock: Union[mp.Lock, mp.RLock], update_dt: float = 1.0):
		"""
		Create a new process to plot the training history.

		:param history_path: The path to the pickle file containing the training history.
		:type history_path: str
		:param lock: The lock to use to protect the training history.
		:type lock: Union[mp.Lock, mp.RLock]
		:param update_dt: The time in seconds between two updates of the plot.
		:type update_dt: float
		"""
		super().__init__()
		self._history_path = history_path
		self._lock = lock
		self._update_dt = update_dt

		self._close_event = mp.Event()
		self._training_history = None
		self.fig, self.axes, self.lines = None, None, None

	def run(self):
		"""
		Run the process.
		
		:return: None
		"""
		can_plot = False
		while not can_plot and not self._close_event.is_set():
			if os.path.exists(self._history_path):
				try:
					with self._lock:
						self._training_history = pickle.load(open(self._history_path, "rb"))
					can_plot = True
				except Exception:
					time.sleep(self._update_dt)

		while len(self._training_history) < 2 and not self._close_event.is_set():
			self.update_history()
			time.sleep(self._update_dt)

		plt.ion()
		self.fig, self.axes, self.lines = self._training_history.create_plot()
		plt.pause(self._update_dt)
		while not self._close_event.is_set():
			self.update_history()
			self.fig, self.axes, self.lines = self._training_history.update_fig(self.fig, self.axes, self.lines)
			plt.pause(self._update_dt)
		plt.close(self.fig)

	def update_history(self):
		"""
		Update the training history by loading it from the pickle file.
		
		:return: None
		"""
		try:
			with self._lock:
				self._training_history = pickle.load(open(self._history_path, "rb"))
		except Exception as e:
			print(f"{e = }")
			pass

	def join(self, *args, **kwargs):
		"""
		Set the close event.
		Join the current process.
		"""
		self._close_event.set()
		plt.close(self.fig)
		return super().join(*args, **kwargs)


class TrainingHistoryVisualizationCallback(BaseCallback):
	"""
	This callback is used to visualize the training history in real time.
	
	:Attributes:
		- **fig** (plt.Figure): The figure to plot.
		- **axes** (plt.Axes): The axes to plot.
		- **lines** (list): The list of lines to plot.
	"""

	def __init__(
			self,
			temp_folder: str = '~/temp/',
			**kwargs
	):
		"""
		Create a new callback to visualize the training history.
		
		:param temp_folder: The folder where to save the training history.
		:type temp_folder: str
		:param kwargs: The keyword arguments to pass to the base callback.
		"""
		super().__init__(**kwargs)
		self._is_open = False
		os.makedirs(temp_folder, exist_ok=True)
		self._temp_path = os.path.join(
			os.path.expanduser(temp_folder), f'training_visualization{time.time_ns()}.pkl'
		)
		self._lock = mp.Lock()
		self._process = _VisualizerProcess(self._temp_path, self._lock)

	def start(self, trainer, **kwargs):
		"""
		Start the process.
		
		:param trainer: The trainer.
		:type trainer: Trainer
		
		:return: None
		"""
		self._process.start()
		self._is_open = True

	def on_iteration_end(self, trainer, **kwargs):
		"""
		Update the training history.
		
		:param trainer: The trainer.
		:type trainer: Trainer
		
		:return: None
		"""
		with self._lock:
			pickle.dump(trainer.training_history, open(self._temp_path, "wb"))

	def __del__(self):
		"""
		Delete the temporary file.
		
		:return: None
		"""
		self.close(None)

	def close(self, trainer, **kwargs):
		"""
		Close the process adn delete the temporary file.
		
		:param trainer: The trainer.
		:type trainer: Trainer
		
		:return: None
		"""
		if self._is_open:
			self._process.join()
			self._process.kill()
			self._process.close()
			self._is_open = False
			with self._lock:
				if os.path.exists(self._temp_path):
					os.remove(self._temp_path)






import enum
import json
import logging
import os
import pprint
import shutil
import warnings
from typing import Any, Dict, Optional, Union

import numpy as np
import torch

from .base_callback import BaseCallback
from ..utils import mapping_update_recursively


class LoadCheckpointMode(enum.Enum):
	"""
	Enum for the different modes of loading a checkpoint.
	
	:Attributes:
		- **BEST_ITR** (int): Indicates that the iteration with the best metric will be loaded.
		- **LAST_ITR** (int): Indicates that the last iteration will be loaded.
	"""
	BEST_ITR = 0
	LAST_ITR = 1
	
	def __str__(self):
		return self.name.lower()
	
	@staticmethod
	def from_str(mode_name: str) -> 'LoadCheckpointMode':
		"""
		Converts a string to a :class:`LoadCheckpointMode` instance.
		
		:param mode_name: The name of the mode.
		:type mode_name: str
		:return: The corresponding :class:`LoadCheckpointMode` instance.
		:rtype: LoadCheckpointMode
		"""
		if mode_name.lower() in ["best", "last"]:
			mode_name = mode_name.upper() + "_ITR"
		assert mode_name.upper() in LoadCheckpointMode.__members__, f"Invalid mode name: {mode_name}"
		return LoadCheckpointMode[mode_name.upper()]


class CheckpointManager(BaseCallback):
	"""
	This class is used to manage and create the checkpoints of a model.

	:Attributes:
		- **checkpoint_folder** (str): The folder to save the checkpoints to.
		- **meta_path_prefix** (str): The prefix to use for the checkpoint's metadata file.
		- **metric** (str): The name of the metric to collect the best checkpoint on.
		- **minimise_metric** (bool): Whether to minimise the metric or maximise it.
		- **curr_best_metric** (float): The current best metric value.
	"""
	
	DEFAULT_PRIORITY = BaseCallback.DEFAULT_LOW_PRIORITY

	SAVE_EXT: str = '.pth'
	SUFFIX_SEP: str = '-'
	CHECKPOINTS_META_SUFFIX: str = 'checkpoints'
	CHECKPOINT_SAVE_PATH_KEY: str = "save_path"
	CHECKPOINT_BEST_KEY: str = "best"
	CHECKPOINT_ITRS_KEY: str = "iterations"
	CHECKPOINT_ITR_KEY: str = "itr"
	CHECKPOINT_METRICS_KEY: str = 'metrics'
	CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"
	CHECKPOINT_STATE_DICT_KEY: str = "model_state_dict"
	CHECKPOINT_TRAINING_HISTORY_KEY: str = "training_history"
	CHECKPOINT_FILE_STRUCT: Dict[str, Union[str, Dict[int, str]]] = {
		CHECKPOINT_BEST_KEY: CHECKPOINT_SAVE_PATH_KEY,
		CHECKPOINT_ITRS_KEY: {0: CHECKPOINT_SAVE_PATH_KEY},
	}
	load_mode_to_suffix: Dict[LoadCheckpointMode, str] = {mode: mode.name for mode in list(LoadCheckpointMode)}

	@staticmethod
	def _replace_trainer_history(trainer, new_history: Any):
		"""
		Replaces the training history in the trainer with the given history.
		:param trainer: The trainer object.
		:param new_history: The new history object.
		:return: None
		"""
		# trainer.callbacks.remove(trainer.training_history)
		# trainer.callbacks.append(new_history)
		# trainer.training_history = new_history
		trainer.training_history.load_checkpoint_state(trainer, new_history)
		trainer.sort_callbacks_()

	@staticmethod
	def get_save_name_from_checkpoints(
			checkpoints_meta: Dict[str, Union[str, Dict[Any, str]]],
			load_checkpoint_mode: LoadCheckpointMode = LoadCheckpointMode.BEST_ITR
	) -> str:
		"""
		Gets the save name from the checkpoint's metadata given the load checkpoint mode.
		
		:param checkpoints_meta: The checkpoint's metadata.
		:type checkpoints_meta: Dict[str, Union[str, Dict[Any, str]]]
		:param load_checkpoint_mode: The load checkpoint mode.
		:type load_checkpoint_mode: LoadCheckpointMode
		
		:return: The save name.
		:rtype: str
		"""
		if load_checkpoint_mode == load_checkpoint_mode.BEST_ITR:
			if CheckpointManager.CHECKPOINT_BEST_KEY in checkpoints_meta:
				return checkpoints_meta[CheckpointManager.CHECKPOINT_BEST_KEY]
			else:
				raise FileNotFoundError(
					f"No best checkpoint found in checkpoints_meta. "
					f"Please use a different load_checkpoint_mode."
				)
		elif load_checkpoint_mode == load_checkpoint_mode.LAST_ITR:
			itr_dict = checkpoints_meta[CheckpointManager.CHECKPOINT_ITRS_KEY]
			last_itr: int = max([int(e) for e in itr_dict])
			return checkpoints_meta[CheckpointManager.CHECKPOINT_ITRS_KEY][str(last_itr)]
		else:
			raise ValueError("Invalid load_checkpoint_mode")

	def __init__(
			self,
			checkpoint_folder: str = "./checkpoints",
			*,
			meta_path_prefix: str = "network",
			metric: str = "val_loss",
			minimise_metric: bool = True,
			save_freq: int = 1,
			save_best_only: bool = False,
			start_save_at: int = 0,
			verbose: bool = False,
			**kwargs
	):
		"""
		Initialises the checkpoint manager.
		
		:param checkpoint_folder: The folder to save the checkpoints to.
		:type checkpoint_folder: str
		:param meta_path_prefix: The prefix to use for the checkpoint's metadata file.
		:type meta_path_prefix: str
		:param metric: The name of the metric to collect the best checkpoint on.
		:type metric: str
		:param minimise_metric: Whether to minimise the metric or maximise it.
		:type minimise_metric: bool
		:param save_freq: The frequency at which to save checkpoints. If set to <= 0, will save at the end of the
							training.
		:type save_freq: int
		:param save_best_only: Whether to only save the best checkpoint. If set to True, the save_freq will be set
		automatically to -1.
		:type save_best_only: bool
		:param start_save_at: The iteration at which to start saving checkpoints.
		:type start_save_at: int
		:param verbose: Whether to print out the trace of the checkpoint manager.
		:type verbose: bool
		:param kwargs: The keyword arguments to pass to the BaseCallback.
		"""
		kwargs.setdefault("save_state", False)
		super().__init__(**kwargs)
		os.makedirs(checkpoint_folder, exist_ok=True)
		self.checkpoint_folder = checkpoint_folder
		self.meta_path_prefix = meta_path_prefix
		self.verbose = verbose

		self.metric = metric
		self.minimise_metric = minimise_metric
		self.save_freq = save_freq
		self.save_best_only = save_best_only
		if self.save_best_only:
			self.save_freq = -1
		self.start_save_at = start_save_at
		self.curr_best_metric = np.inf if self.minimise_metric else -np.inf
		self.curr_checkpoint = None

	@property
	def checkpoints_meta_path(self) -> str:
		"""
		Gets the path to the checkpoints metadata file.
		
		:return: The path to the checkpoints metadata file.
		:rtype: str
		"""
		full_filename = (
			f"{self.meta_path_prefix}{CheckpointManager.SUFFIX_SEP}{CheckpointManager.CHECKPOINTS_META_SUFFIX}"
		)
		return f"{self.checkpoint_folder}/{full_filename}.json"

	def get_checkpoint_filename(self, itr: int = -1):
		"""
		Generate the filename for the checkpoint at the given iteration.
		
		:param itr: The iteration to generate the filename for.
		:type itr: int
		
		:return: The filename for the checkpoint at the given iteration.
		:rtype: str
		"""
		pre_name = f"{self.meta_path_prefix}"
		if itr == -1:
			post_name = ""
		else:
			post_name = f"{CheckpointManager.SUFFIX_SEP}{CheckpointManager.CHECKPOINT_ITR_KEY}{itr}"
		return f"{pre_name}{post_name}{CheckpointManager.SAVE_EXT}"

	def _create_new_checkpoint_meta(self, itr: int, best: bool = False) -> dict:
		"""
		Creates a new checkpoint's metadata.
		
		:param itr: The iteration of the checkpoint.
		:type itr: int
		:param best: Whether the checkpoint is currently the best.
		:type best: bool
		
		:return: The new checkpoint's metadata.
		:rtype: dict
		"""
		save_name = self.get_checkpoint_filename(itr)
		new_info = {CheckpointManager.CHECKPOINT_ITRS_KEY: {str(itr): save_name}}
		if best:
			new_info[CheckpointManager.CHECKPOINT_BEST_KEY] = save_name
		return new_info

	def save_checkpoint(
			self,
			itr: int,
			itr_metrics: Dict[str, Any],
			best: bool = False,
			state_dict: Optional[Dict[str, Any]] = None,
			optimizer_state_dict: Optional[Dict[str, Any]] = None,
			training_history: Optional[Any] = None,
			**other_states,
	) -> str:
		"""
		Saves a checkpoint of the model and optimizer states at the given iteration.
		
		:param itr: The iteration number.
		:type itr: int
		:param itr_metrics: The metrics at the given iteration.
		:type itr_metrics: Dict[str, Any]
		:param best: Whether this is the best iteration so far.
		:type best: bool
		:param state_dict: The state dict of the model.
		:type state_dict: Optional[Dict[str, Any]]
		:param optimizer_state_dict: The state dict of the optimizer.
		:type optimizer_state_dict: Optional[Dict[str, Any]]
		:param training_history: The training history object.
		:type training_history: Optional[Any]
		
		:return: The path to the saved checkpoint.
		:rtype: str
		"""
		os.makedirs(self.checkpoint_folder, exist_ok=True)
		save_name = self.get_checkpoint_filename(itr)
		path = os.path.join(self.checkpoint_folder, save_name)
		basic_states = {
			CheckpointManager.CHECKPOINT_ITR_KEY: itr,
			CheckpointManager.CHECKPOINT_STATE_DICT_KEY: state_dict,
			CheckpointManager.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: optimizer_state_dict,
			CheckpointManager.CHECKPOINT_METRICS_KEY: itr_metrics,
			CheckpointManager.CHECKPOINT_TRAINING_HISTORY_KEY: training_history,
		}
		assert all(key not in other_states for key in basic_states.keys()), (
			f"Other states cannot have the same keys as the basic states ({basic_states.keys()})."
		)
		states = {**basic_states, **other_states}
		torch.save(states, path)
		self.save_checkpoints_meta(self._create_new_checkpoint_meta(itr, best))
		return path

	def load_checkpoint(
			self,
			load_checkpoint_mode: LoadCheckpointMode = LoadCheckpointMode.BEST_ITR
	) -> dict:
		"""
		Loads the checkpoint at the given load_checkpoint_mode.
		
		:param load_checkpoint_mode: The load_checkpoint_mode to use.
		:type load_checkpoint_mode: LoadCheckpointMode
		
		:return: The loaded checkpoint.
		:rtype: dict
		"""
		# TODO: add the possibility to load a specific itr
		with open(self.checkpoints_meta_path, "r+") as jsonFile:
			info: dict = json.load(jsonFile)
		filename = CheckpointManager.get_save_name_from_checkpoints(info, load_checkpoint_mode)
		checkpoint = torch.load(f"{self.checkpoint_folder}/{filename}")
		return checkpoint

	def save_checkpoints_meta(self, new_info: dict):
		"""
		Saves the new checkpoints' metadata.
		
		:param new_info: The new checkpoints' metadata.
		:type new_info: dict
		
		:return: None
		"""
		info = dict()
		if os.path.exists(self.checkpoints_meta_path):
			with open(self.checkpoints_meta_path, "r+") as jsonFile:
				info = json.load(jsonFile)
		mapping_update_recursively(info, new_info)
		os.makedirs(self.checkpoint_folder, exist_ok=True)
		with open(self.checkpoints_meta_path, "w+") as jsonFile:
			json.dump(info, jsonFile, indent=4)

	def start(self, trainer, **kwargs):
		"""
		Call at the beginning of the training by the Trainer. Load the checkpoint base on the load_checkpoint_mode of
		the trainer and update the current_training_state of the trainer.
		
		:param trainer: The trainer.
		:type trainer: Trainer
		
		:return: None
		"""
		super().start(trainer)
		if trainer.load_checkpoint_mode is None:
			trainer.load_checkpoint_mode = LoadCheckpointMode.LAST_ITR
		start_itr = 0
		checkpoint = self.curr_checkpoint
		if trainer.force_overwrite:
			if os.path.exists(self.checkpoints_meta_path):
				shutil.rmtree(self.checkpoint_folder)
		else:
			try:
				checkpoint = self.load_checkpoint(trainer.load_checkpoint_mode)
				trainer.model.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_STATE_DICT_KEY], strict=True)
				if trainer.optimizer is not None:
					trainer.optimizer.load_state_dict(checkpoint[CheckpointManager.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY])
				start_itr = int(checkpoint[CheckpointManager.CHECKPOINT_ITR_KEY]) + 1
				# self._replace_trainer_history(trainer, checkpoint[CheckpointManager.CHECKPOINT_TRAINING_HISTORY_KEY])
			except FileNotFoundError as e:
				if self.verbose:
					logging.info("No such checkpoint. Fit from beginning.")
			finally:
				self.curr_checkpoint = checkpoint

		trainer.update_state_(iteration=start_itr)
		if self.minimise_metric:
			self.curr_best_metric = trainer.training_history.min(self.metric)
		else:
			self.curr_best_metric = trainer.training_history.max(self.metric)
	
	def save_on(self, trainer) -> bool:
		"""
		Saves the checkpoint if the current iteration is a checkpoint iteration.
		
		:param trainer: The trainer.
		:type trainer: Trainer
		
		:return: Whether the checkpoint was saved.
		:rtype: bool
		"""
		itr_metrics = trainer.current_training_state.itr_metrics
		if itr_metrics is None:
			itr_metrics = {}
		other_states = trainer.callbacks.get_checkpoint_state(trainer)
		is_best = self._check_is_best(trainer)
		if is_best:
			self.curr_best_metric = itr_metrics[self.metric]
		self.save_checkpoint(
			trainer.current_training_state.iteration, itr_metrics, is_best,
			state_dict=trainer.model.state_dict(),
			optimizer_state_dict=trainer.optimizer.state_dict() if trainer.optimizer else None,
			training_history=trainer.training_history.get_checkpoint_state(trainer),
			**other_states
		)
		if trainer.training_history:
			trainer.training_history.plot(
				save_path=os.path.join(self.checkpoint_folder, "training_history.png"),
				show=False
			)
		return True

	def on_iteration_end(self, trainer, **kwargs):
		"""
		Called when an iteration ends. An iteration is defined as one full pass through the training dataset and
		the validation dataset. The checkpoint is saved if the current constraints are met.
		
		:param trainer: The trainer.
		:type trainer: Trainer
		
		:return: None
		"""
		if self.start_save_at > trainer.current_training_state.iteration:
			return
		if self.save_best_only:
			if self._check_is_best(trainer):
				self.save_on(trainer)
		elif self.save_freq > 0 and trainer.current_training_state.iteration % self.save_freq == 0:
			self.save_on(trainer)
		if trainer.current_training_state.iteration >= trainer.state.n_iterations - 1:
			self.save_on(trainer)
	
	def _check_is_best(self, trainer) -> Optional[bool]:
		if trainer.current_training_state.itr_metrics is None:
			return None
		
		if self.minimise_metric:
			is_best = trainer.current_training_state.itr_metrics[self.metric] < self.curr_best_metric
		else:
			is_best = trainer.current_training_state.itr_metrics[self.metric] > self.curr_best_metric
		return is_best
	
	def close(self, trainer, **kwargs):
		"""
		Called when the training is finished. Saves the current checkpoint if the current iteration is lower than
		the number of iterations i.e. there is new stuff to save.
		
		:param trainer: The trainer.
		:type trainer: Trainer
		
		:return: None
		"""
		if trainer.current_training_state.iteration < trainer.state.n_iterations:
			self.save_on(trainer)



import warnings
from collections import OrderedDict
from copy import deepcopy, copy
from typing import Iterable, Optional, List, Callable, Dict, Any, Union, NamedTuple, Generator

import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from ..transforms.base import to_numpy, ToTensor
from ..learning_algorithms.bptt import BPTT
from ..callbacks import CheckpointManager, LoadCheckpointMode, TrainingHistory
from ..callbacks.base_callback import BaseCallback, CallbacksList
from ..learning_algorithms.learning_algorithm import LearningAlgorithm
from ..modules import BaseModel
from ..regularization import BaseRegularization, RegularizationList


class CurrentTrainingState(NamedTuple):
	r"""
	This class is used to store the current training state. It is extremely useful for the callbacks
	to access the current training state and to personalize the training process.
	
	:Attributes:
		- **n_iterations** (int): The total number of iterations.
		- **iteration** (int): The current iteration.
		- **epoch** (int): The current epoch.
		- **batch** (int): The current batch.
		- **x_batch** (Any): The current input batch.
		- **y_batch** (Any): The current target batch.
		- **pred_batch** (Any): The current prediction.
		- **batch_loss** (float): The current loss.
		- **batch_is_train** (bool): Whether the current batch is a training batch.
		- **train_loss** (float): The current training loss.
		- **val_loss** (float): The current validation loss.
		- **itr_metrics** (Dict[str, Any]): The current iteration metrics.
		- **stop_training_flag** (bool): Whether the training should be stopped.
		- **info** (Dict[str, Any]): Any additional information. This is useful to communicate between callbacks.
		- **objects** (Dict[str, Any]): Any additional objects. This is useful to manage objects between callbacks.
			Note: In general, the train_dataloader and val_dataloader should be stored here.
	
	"""
	n_iterations: Optional[int] = None
	iteration: Optional[int] = None
	n_epochs: Optional[int] = None
	epoch: Optional[int] = None
	epoch_loss: Optional[Any] = None
	batch: Optional[int] = None
	x_batch: Optional[Any] = None
	y_batch: Optional[Any] = None
	pred_batch: Optional[Any] = None
	batch_loss: Optional[Any] = None
	batch_is_train: Optional[bool] = None
	train_loss: Optional[Any] = None
	val_loss: Optional[Any] = None
	train_metrics: Optional[Any] = None
	val_metrics: Optional[Any] = None
	itr_metrics: Optional[Dict[str, Any]] = {}
	stop_training_flag: bool = False
	info: Dict[str, Any] = {}
	objects: Dict[str, Any] = {}
	
	def __getstate__(self):
		not_picklable = ["info", "objects"]
		d = {k: v for k, v in self._asdict().items() if k not in not_picklable}
		return d
	
	@staticmethod
	def get_null_state() -> "CurrentTrainingState":
		return CurrentTrainingState()
	
	def update(self, **kwargs) -> "CurrentTrainingState":
		self_dict = self._asdict()
		assert all(k in self_dict for k in kwargs)
		self_dict.update(kwargs)
		return CurrentTrainingState(**self_dict)
	

TrainingState = CurrentTrainingState  # Alias


class Trainer:
	def __init__(
			self,
			model: BaseModel,
			*,
			predict_method: str = "__call__",
			criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] = None,
			regularization: Optional[Union[BaseRegularization, RegularizationList, Iterable[BaseRegularization]]] = None,
			optimizer: Optional[torch.optim.Optimizer] = None,
			learning_algorithm: Optional[LearningAlgorithm] = None,
			regularization_optimizer: Optional[torch.optim.Optimizer] = None,
			metrics: Optional[List[Callable]] = None,
			callbacks: Optional[Union[List[BaseCallback], CallbacksList, BaseCallback]] = None,
			device: Optional[torch.device] = None,
			verbose: bool = True,
			**kwargs
	):
		"""
		Constructor for Trainer.

		:param model: Model to train.
		:param criterion: Loss function(s) to use. Deprecated, use `learning_algorithm` instead.
		:param regularization: Regularization(s) to use. In NeuroTorch, there are two ways to do regularization:
			1. Regularization can be specified in the layers with the 'update_regularization_loss' method. This
			regularization will be performed by the same optimizer as the main loss. This way is useful when you
			want a regularization that depends on the model output or hidden state.
			2. Regularization can be specified in the trainer with the 'regularization' parameter. This regularization
			will be performed by a separate optimizer named 'regularization_optimizer'. This way is useful when you
			want a regularization that depends only on the model parameters and when you want to control the
			learning rate of the regularization independently of the main loss.
			
			Note: This parameter will be deprecated and remove in a future version. The regularization will be
				specified in the learning algorithm and/or in the callbacks.
			
		:param optimizer: Optimizer to use for the main loss. Deprecated. Use learning_algorithm instead.
		:param learning_algorithm: Learning algorithm to use for the main loss. This learning algorithm can be given
			in the callbacks list as well. If specified, this learning algorithm will be added to the callbacks list.
			In this case, make sure that the learning algorithm is not added twice. Note that multiple learning
			algorithms can be used in the callbacks list.
		:param regularization_optimizer: Optimizer to use for the regularization loss.
		:param metrics: Metrics to compute during training.
		:param callbacks: Callbacks to use during training. Each callback will be called at different moments,
			see the documentation of :class:`BaseCallback` for more information.
		:param device: Device to use for the training. Default is the device of the model.
		:param verbose: Whether to print information during training.
		:param kwargs: Additional arguments of the training.
		
		:keyword int n_epochs: The number of epochs to train at each iteration. Default is 1.
		:keyword float lr: Learning rate of the main optimizer. Default is 1e-3.
		:keyword float reg_lr: Learning rate of the regularization optimizer. Default is 1e-2.
		:keyword float weight_decay: Weight decay of the main optimizer. Default is 0.0.
		:keyword bool exec_metrics_on_train: Whether to compute metrics on the train dataset. This is useful when
			you want to save time by not computing the metrics on the train dataset. Default is True.
		:keyword x_transform: Transform to apply to the input data before passing it to the model.
		:keyword y_transform: Transform to apply to the target data before passing it to the model. For example,
			this can be used to convert the target data to a one-hot encoding or to long tensor
			using `nt.ToTensor(dtype=torch.long)`.
		"""
		assert model.is_built, "Model must be built before training"
		self.kwargs = self._set_default_kwargs(kwargs)
		self.model = model
		self.predict_method = predict_method
		assert hasattr(model, predict_method), f"Model does not have a method named '{predict_method}'"
		assert callable(getattr(model, predict_method)), f"Model method '{predict_method}' is not callable"
		self.criterion = self._set_default_criterion(criterion)
		self.regularization = self._set_default_regularization(regularization)
		# self._maybe_add_regularization(self.regularization)
		self.optimizer = self._set_default_optimizer(optimizer)
		self.regularization_optimizer = self._set_default_reg_optimizer(regularization_optimizer)
		self.metrics = self._set_default_metrics(metrics)
		self.callbacks: CallbacksList = self._set_default_callbacks(callbacks)
		self._maybe_add_learning_algorithm(learning_algorithm)
		self.sort_callbacks_()
		self.device = self._set_default_device(device)
		self.verbose = verbose
		self.training_history: TrainingHistory = self.training_histories[0]
		self.current_training_state = CurrentTrainingState()
		
		self.x_transform = kwargs.get("x_transform", ToTensor())
		self.y_transform = kwargs.get("y_transform", ToTensor())

		self._load_checkpoint_mode = None
		self._force_overwrite = None
	
	@property
	def network(self):
		"""
		Alias for the model.
		
		:return: The :attr:`model` attribute.
		"""
		return self.model
	
	@network.setter
	def network(self, value):
		"""
		Alias for the model.
		
		:param value: The new value for the :attr:`model` attribute.
		:return: None
		"""
		self.model = value
	
	@property
	def state(self):
		"""
		Alias for the :attr:`current_training_state` attribute.
		
		:return: The :attr:`current_training_state`
		"""
		return self.current_training_state

	@property
	def load_checkpoint_mode(self):
		return self._load_checkpoint_mode
	
	@load_checkpoint_mode.setter
	def load_checkpoint_mode(self, value: LoadCheckpointMode):
		self._load_checkpoint_mode = value

	@property
	def force_overwrite(self):
		return self._force_overwrite

	@property
	def training_histories(self) -> CallbacksList:
		return CallbacksList(list(filter(lambda x: isinstance(x, TrainingHistory), self.callbacks)))
	
	@property
	def checkpoint_managers(self) -> CallbacksList:
		return CallbacksList(list(filter(lambda x: isinstance(x, CheckpointManager), self.callbacks)))
	
	@property
	def learning_algorithms(self) -> CallbacksList:
		return CallbacksList(list(filter(lambda x: isinstance(x, LearningAlgorithm), self.callbacks)))
	
	@staticmethod
	def _set_default_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
		kwargs.setdefault("n_epochs", 1)
		kwargs.setdefault("lr", 1e-3)
		kwargs.setdefault("reg_lr", 1e-2)
		kwargs.setdefault("weight_decay", 0.0)
		kwargs.setdefault("batch_size", 256)
		kwargs.setdefault("exec_metrics_on_train", True)
		
		assert kwargs["batch_size"] > 0, "batch_size must be positive"
		return kwargs
	
	def _set_default_optimizer(self, optimizer: Optional[torch.optim.Optimizer]) -> torch.optim.Optimizer:
		warnings.warn("The 'optimizer' parameter is deprecated. Use the 'callbacks' parameter instead.", DeprecationWarning)
		# if optimizer is None:
		# 	optimizer = torch.optim.Adam(
		# 		self.model.parameters(),
		# 		lr=self.kwargs["lr"],
		# 		weight_decay=self.kwargs["weight_decay"],
		# 	)
		return optimizer
	
	def _maybe_add_learning_algorithm(self, learning_algorithm: Optional[LearningAlgorithm]) -> None:
		if len(self.learning_algorithms) == 0 and learning_algorithm is None:
			learning_algorithm = BPTT(optimizer=self.optimizer, criterion=self.criterion)
		if learning_algorithm is not None:
			self.callbacks.append(learning_algorithm)
			
	def _maybe_add_regularization(self, regularization: Optional[RegularizationList]) -> None:
		if regularization is not None:
			self.callbacks.append(regularization)
	
	def _set_default_reg_optimizer(self, optimizer: Optional[torch.optim.Optimizer]) -> torch.optim.Optimizer:
		warnings.warn("The 'regularization_optimizer' parameter is deprecated. Use the 'callbacks' parameter instead.", DeprecationWarning)
		if optimizer is None and self.regularization is not None:
			optimizer = torch.optim.SGD(
				self.regularization.parameters(),
				lr=self.kwargs["reg_lr"],
				weight_decay=0.0,
			)
		return optimizer

	def _set_default_metrics(self, metrics: Optional[List[Callable]]):
		if metrics is None:
			metrics = []
		return metrics
	
	def _set_default_criterion(self, criterion: Optional[torch.nn.Module]) -> torch.nn.Module:
		warnings.warn("The 'criterion' parameter is deprecated. Use the 'callbacks' parameter instead.", DeprecationWarning)
		# if criterion is None:
		# 	if isinstance(self.model.output_sizes, dict):
		# 		criterion = {
		# 			k: torch.nn.MSELoss() for k in self.model.output_sizes
		# 		}
		# 	elif isinstance(self.model.output_sizes, int):
		# 		criterion = torch.nn.MSELoss()
		# 	else:
		# 		raise ValueError("Unknown criterion type")
		return criterion

	def _set_default_regularization(
			self,
			regularization: Optional[Union[BaseRegularization, RegularizationList, Iterable[BaseRegularization]]]
	) -> Optional[RegularizationList]:
		warnings.warn("The 'regularization' parameter is deprecated. Use the 'callbacks' parameter instead.", DeprecationWarning)
		if regularization is None:
			pass
		elif isinstance(regularization, BaseRegularization):
			regularization = RegularizationList([regularization])
		elif isinstance(regularization, RegularizationList):
			pass
		elif isinstance(regularization, Iterable):
			regularization = RegularizationList(regularization)
		return regularization
	
	def _set_default_device(self, device: Optional[torch.device]) -> torch.device:
		if device is None:
			device = self.model.device
		return device
	
	@staticmethod
	def _set_default_callbacks(
			callbacks: Optional[Union[List[BaseCallback], CallbacksList, BaseCallback]]
	) -> CallbacksList:
		if callbacks is None:
			callbacks = []
		if isinstance(callbacks, BaseCallback):
			callbacks = [callbacks]
		if not any([isinstance(callback, TrainingHistory) for callback in callbacks]):
			callbacks.append(TrainingHistory())
		return CallbacksList(callbacks)
	
	def update_state_(self, **kwargs):
		self.current_training_state = self.current_training_state.update(**kwargs)
	
	def update_objects_state_(self, **kwargs):
		self.update_state_(objects={**self.current_training_state.objects, **kwargs})
	
	def update_info_state_(self, **kwargs):
		self.update_state_(info={**self.current_training_state.info, **kwargs})
		
	def update_itr_metrics_state_(self, **kwargs):
		self.update_state_(itr_metrics={**self.current_training_state.itr_metrics, **kwargs})
	
	def sort_callbacks_(self, reverse: bool = False) -> CallbacksList:
		"""
		Sort the callbacks by their priority. The higher the priority, the earlier the callback is called. In general,
		the callbacks will be sorted in the following order:
			1. TrainingHistory callbacks;
			2. Others callbacks;
			3. CheckpointManager callbacks.
		
		:param reverse: Whether to reverse the order of the callbacks. Default is False.
		:type reverse: bool
		:return: The sorted callbacks.
		:rtype: CallbacksList
		"""
		self.callbacks.sort_callbacks_(reverse=reverse)
		return self.callbacks
	
	def load_state(self):
		"""
		Load the state of the trainer from the checkpoint.
		"""
		if self.checkpoint_managers:
			main_checkpoint_manager: CheckpointManager = self.checkpoint_managers[0]
			checkpoint = main_checkpoint_manager.curr_checkpoint
			if checkpoint:
				self.callbacks.load_checkpoint_state(self, checkpoint)

	def train(
			self,
			train_dataloader: DataLoader,
			val_dataloader: Optional[DataLoader] = None,
			n_iterations: Optional[int] = None,
			*,
			n_epochs: int = 1,
			load_checkpoint_mode: LoadCheckpointMode = None,
			force_overwrite: bool = False,
			p_bar_position: Optional[int] = None,
			p_bar_leave: Optional[bool] = None,
			**kwargs
	) -> TrainingHistory:
		"""
		Train the model.
		
		:param train_dataloader: The dataloader for the training set. It contains the training data.
		:type train_dataloader: DataLoader
		:param val_dataloader: The dataloader for the validation set. It contains the validation data.
		:type val_dataloader: Optional[DataLoader]
		:param n_iterations: The number of iterations to train the model. An iteration is a pass over the training set
			and the validation set. If None, the model will be trained until the training is stopped by the user.
		:type n_iterations: Optional[int]
		:param n_epochs: The number of epochs to train the model. An epoch is a pass over the training set. The
			nomenclature here is different from what is usually used elsewhere. Here, an epoch is a pass over the
			training set, while an iteration is a pass over the training set and the validation set. In other words,
			if `n_iterations=1` and `n_epochs=10`, the trainer will pass 10 times over the training set  and 1 time
			over the validation set (this will constitute 1 iteration). If `n_iterations=10` and `n_epochs=1`, the
			trainer will pass 10 times over the training set and 10 times over the validation set (this will constitute
			10 iterations). The nuance between those terms is really important when is comes to reinforcement learning.
			Default is 1.
		:type n_epochs: int
		:param load_checkpoint_mode: The mode to use when loading the checkpoint.
		:type load_checkpoint_mode: LoadCheckpointMode
		:param force_overwrite: Whether to force overwriting the checkpoint. Be careful when using this option, as it
			will destroy the previous checkpoint folder. Default is False.
		:type force_overwrite: bool
		:param p_bar_position: The position of the progress bar. See tqdm documentation for more information.
		:type p_bar_position: Optional[int]
		:param p_bar_leave: Whether to leave the progress bar. See tqdm documentation for more information.
		:type p_bar_leave: Optional[bool]
		:param kwargs: Additional keyword arguments.
		
		:return: The training history.
		"""
		self._load_checkpoint_mode = load_checkpoint_mode
		self._force_overwrite = force_overwrite
		self.kwargs.update(kwargs)
		self.update_state_(
			n_iterations=n_iterations,
			n_epochs=n_epochs,
			objects={
				**self.current_training_state.objects,
				**{"train_dataloader": train_dataloader, "val_dataloader": val_dataloader}
			}
		)
		self.sort_callbacks_()
		self.callbacks.start(self)
		self.load_state()
		if self.current_training_state.iteration is None:
			self.update_state_(iteration=0)
		if len(self.training_history) > 0:
			self.update_itr_metrics_state_(**self.training_history.get_item_at(-1))
		else:
			self.update_state_(itr_metrics={})
		p_bar = tqdm(
			initial=self.current_training_state.iteration,
			total=self.current_training_state.n_iterations,
			desc=kwargs.get("desc", "Training"),
			disable=not self.verbose,
			position=p_bar_position,
			unit="itr",
			leave=p_bar_leave
		)
		for i in self._iterations_generator(p_bar):
			self.update_state_(iteration=i)
			self.callbacks.on_iteration_begin(self)
			train_dataloader = self.current_training_state.objects["train_dataloader"]
			val_dataloader = self.current_training_state.objects["val_dataloader"]
			itr_loss = self._exec_iteration(train_dataloader, val_dataloader)
			if self.kwargs["exec_metrics_on_train"]:
				itr_train_metrics = self._exec_metrics(train_dataloader, prefix="train")
			else:
				itr_train_metrics = {}
			if val_dataloader is not None:
				itr_val_metrics = self._exec_metrics(val_dataloader, prefix="val")
			else:
				itr_val_metrics = {}
			self.update_itr_metrics_state_(**dict(**itr_loss, **itr_train_metrics, **itr_val_metrics))
			postfix = {f"{k}": f"{v:.5e}" for k, v in self.state.itr_metrics.items()}
			postfix.update(self.callbacks.on_pbar_update(self))
			self.callbacks.on_iteration_end(self)
			p_bar.set_postfix(postfix)
			if self.current_training_state.stop_training_flag:
				p_bar.set_postfix(OrderedDict(**{"stop_flag": "True"}, **postfix))
				break
		self.callbacks.close(self)
		p_bar.close()
		return self.training_history
	
	def _iterations_generator(self, p_bar: tqdm) -> Generator:
		"""
		Generator that yields the current iteration and updates the state and the p_bar.
		
		:return: The current iteration.
		"""
		while self.current_training_state.iteration < self._get_numeric_n_iterations():
			yield self.current_training_state.iteration
			self.update_state_(iteration=self.current_training_state.iteration + 1)
			p_bar.total = self.current_training_state.n_iterations
			p_bar.update()
	
	def _get_numeric_n_iterations(self) -> int:
		"""
		Returns the number of iterations.
		
		:return: The number of iterations.
		"""
		n_iterations = self.current_training_state.n_iterations
		if n_iterations is None:
			n_iterations = np.inf
		return n_iterations

	def _exec_iteration(
			self,
			train_dataloader: DataLoader,
			val_dataloader: Optional[DataLoader] = None
	) -> Dict[str, float]:
		with torch.no_grad():
			torch.cuda.empty_cache()
		losses = {}
		
		self.model.train()
		self.callbacks.on_train_begin(self)
		self.update_state_(batch_is_train=True)
		train_losses = []
		for epoch_idx in range(self.current_training_state.n_epochs):
			self.update_state_(epoch=epoch_idx)
			train_losses.append(self._exec_epoch(train_dataloader))
		train_loss = np.mean(train_losses)
		self.update_state_(train_loss=train_loss)
		self.callbacks.on_train_end(self)
		losses["train_loss"] = train_loss

		if val_dataloader is not None:
			with torch.no_grad():
				self.model.eval()
				self.callbacks.on_validation_begin(self)
				self.update_state_(batch_is_train=False)
				val_loss = self._exec_epoch(val_dataloader)
				self.update_state_(val_loss=val_loss)
				self.callbacks.on_validation_end(self)
				losses["val_loss"] = val_loss
		
		with torch.no_grad():
			torch.cuda.empty_cache()
		return losses

	def _exec_metrics(self, dataloader: torch.utils.data.DataLoader, prefix: str) -> Dict:
		metrics_dict = {}
		for metric in self.metrics:
			m_out = metric(dataloader)
			if isinstance(m_out, dict):
				metrics_dict.update({f"{prefix}_{k}": v for k, v in m_out.items()})
			else:
				metric_name = str(metric)
				if hasattr(metric, "name"):
					metric_name = metric.name
				elif hasattr(metric, "__name__"):
					metric_name = metric.__name__
				metrics_dict[f"{prefix}_{metric_name}"] = m_out
		return metrics_dict

	def _exec_epoch(
			self,
			dataloader: DataLoader,
	) -> float:
		self.callbacks.on_epoch_begin(self)
		batch_losses = []
		for i, (x_batch, y_batch) in enumerate(dataloader):
			self.update_state_(batch=i)
			batch_losses.append(to_numpy(self._exec_batch(x_batch, y_batch)))
		mean_loss = np.mean(batch_losses)
		self.callbacks.on_epoch_end(self)
		return mean_loss

	def _exec_batch(
			self,
			x_batch,
			y_batch,
	):
		x_batch = self.x_transform(self._batch_to_dense(self._batch_to_device(x_batch)))
		y_batch = self.y_transform(self._batch_to_dense(self._batch_to_device(y_batch)))
		self.update_state_(x_batch=x_batch, y_batch=y_batch)
		self.callbacks.on_batch_begin(self)
		pred_batch = self.get_pred_batch(x_batch)
		self.update_state_(pred_batch=pred_batch)
		if self.model.training:
			self.callbacks.on_optimization_begin(self, x=x_batch, y=y_batch, pred=pred_batch)
			self.callbacks.on_optimization_end(self)
		else:
			self.callbacks.on_validation_batch_begin(self, x=x_batch, y=y_batch, pred=pred_batch)
			self.callbacks.on_validation_batch_end(self)
		self.callbacks.on_batch_end(self)
		batch_loss = self.current_training_state.batch_loss
		if batch_loss is None:
			batch_loss = 0.0
		else:
			batch_loss = batch_loss.item()
		return batch_loss
	
	def get_pred_batch(
			self,
			x_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
	):
		if self.model.training:
			out = getattr(self.model, self.predict_method)(x_batch)
		else:
			with torch.no_grad():
				out = getattr(self.model, self.predict_method)(x_batch)
		return out
	
	def format_pred_batch(
			self,
			raw_pred_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
			y_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
	):
		if isinstance(raw_pred_batch, (tuple, list)):
			pred_batch = raw_pred_batch[0]
		elif isinstance(raw_pred_batch, torch.Tensor):
			pred_batch = raw_pred_batch
		else:
			raise ValueError(f"Unsupported output type: {type(raw_pred_batch)}")
		
		if isinstance(y_batch, dict):
			assert isinstance(pred_batch, dict) and isinstance(y_batch, dict), \
				"If y_batch is a dict, pred must be a dict too."
			assert set(pred_batch.keys()) == set(y_batch.keys()), \
				"Keys of y_batch and pred_batch must be the same."
		else:
			if isinstance(pred_batch, dict):
				assert len(pred_batch) == 1, \
					"pred_batch must have only one key if y_batch is not a dict."
				pred_batch = pred_batch[list(pred_batch.keys())[0]]
		return pred_batch

	def apply_criterion_on_batch(
			self,
			x_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
			y_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
			pred_batch: Optional[Union[torch.Tensor, Dict[str, torch.Tensor]]]
	) -> torch.Tensor:
		if isinstance(pred_batch, (tuple, list)):
			pred = pred_batch[0]
		elif isinstance(pred_batch, torch.Tensor):
			pred = pred_batch
		else:
			raise ValueError(f"Unsupported output type: {type(pred_batch)}")
		
		if isinstance(self.criterion, dict):
			if isinstance(y_batch, torch.Tensor):
				y_batch = {k: y_batch for k in self.criterion}
			if isinstance(pred, torch.Tensor):
				pred = {k: pred for k in self.criterion}
			assert isinstance(pred, dict) and isinstance(y_batch, dict) and isinstance(pred, dict), \
				"If criterion is a dict, pred, y_batch and pred must be a dict too."
			batch_loss = sum([
				self.criterion[k](pred[k], y_batch[k].to(self.device))
				for k in self.criterion
			])
		else:
			if isinstance(pred, dict) and len(pred) == 1:
				pred = pred[list(pred.keys())[0]]
			batch_loss = self.criterion(pred, y_batch.to(self.device))
		return batch_loss

	def _batch_to_dense(self, batch):
		if isinstance(batch, dict):
			return {k: self._batch_to_dense(v) for k, v in batch.items()}
		if isinstance(batch, torch.Tensor) and batch.is_sparse:
			return batch.to_dense()
		return batch

	def _batch_to_device(self, batch):
		if isinstance(batch, dict):
			return {k: self._batch_to_device(v) for k, v in batch.items()}
		if isinstance(batch, torch.Tensor):
			return batch.to(self.device, non_blocking=True)
		return batch
	
	def __repr__(self):
		repr_ = f"{self.__class__.__name__}("
		repr_ += f"\n\tmodel={self.model}, "
		repr_ += f"\n\tpredict_method={self.predict_method}, "
		repr_ += f"\n\tcallbacks={self.callbacks}"
		repr_ += f")@{self.device}"
		return repr_
		





from typing import Optional

import numpy as np

from .base_callback import BaseCallback


class EarlyStopping(BaseCallback):
	def __init__(
			self,
			patience: int = 5,
			tol: float = 1e-2,
	):
		self.patience = patience
		self.tol = tol
		self.best_loss = None
		self.counter = 0
	
	def _check_early_stopping(self, patience: int, tol: float = 1e-2) -> bool:
		"""
		:param patience:
		:return:
		"""
		losses = self.loss_history['val'][-patience:]
		return np.all(np.abs(np.diff(losses)) < tol)
		

class EarlyStoppingThreshold(BaseCallback):
	"""
	Monitor the training process and set the stop_training_flag to True when the threshold is met.
	"""
	def __init__(
			self,
			*,
			metric: str,
			threshold: float,
			minimize_metric: bool,
			**kwargs
	):
		"""
		Constructor for EarlyStoppingThreshold class.
		
		:param metric: Name of the metric to monitor.
		:type metric: str
		:param threshold: Threshold value for the metric.
		:type threshold: float
		:param minimize_metric: Whether to minimize or maximize the metric.
		:type minimize_metric: bool
		:param kwargs: The keyword arguments to pass to the BaseCallback.
		"""
		super().__init__(**kwargs)
		self.threshold = threshold
		self.metric = metric
		self.minimize_metric = minimize_metric
	
	def on_iteration_end(self, trainer, **kwargs):
		if self.minimize_metric:
			threshold_met = trainer.current_training_state.itr_metrics[self.metric] < self.threshold
		else:
			threshold_met = trainer.current_training_state.itr_metrics[self.metric] > self.threshold
		if threshold_met:
			trainer.update_state_(stop_training_flag=True)



import os
from collections import defaultdict
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np

from .curriculum import Curriculum
from ..callbacks import TrainingHistory
from ..utils import legend_without_duplicate_labels_


class TrainingHistoriesMap:
	REPORT_KEY = "report"

	def __init__(self, curriculum: Optional[Curriculum] = None):
		self.curriculum = curriculum
		self.histories = defaultdict(TrainingHistory, **{TrainingHistoriesMap.REPORT_KEY: TrainingHistory()})

	@property
	def report_history(self) -> TrainingHistory:
		return self.histories[TrainingHistoriesMap.REPORT_KEY]

	def max(self, key=None):
		if self.curriculum is None:
			return self.histories[TrainingHistoriesMap.REPORT_KEY].max(key)
		else:
			return self.histories[self.curriculum.current_lesson.name].max(key)

	def concat(self, other):
		self.histories[TrainingHistoriesMap.REPORT_KEY].concat(other)
		if self.curriculum is not None:
			return self.histories[self.curriculum.current_lesson.name].concat(other)

	def append(self, key, value):
		self.histories[TrainingHistoriesMap.REPORT_KEY].append(key, value)
		if self.curriculum is not None:
			return self.histories[self.curriculum.current_lesson.name].append(key, value)

	@staticmethod
	def _set_default_plot_kwargs(kwargs: dict):
		kwargs.setdefault('fontsize', 16)
		kwargs.setdefault('linewidth', 3)
		kwargs.setdefault('figsize', (16, 12))
		kwargs.setdefault('dpi', 300)
		return kwargs

	def plot(self, save_path=None, show=False, lesson_idx: Optional[Union[int, str]] = None, **kwargs):
		kwargs = self._set_default_plot_kwargs(kwargs)
		if self.curriculum is None:
			assert lesson_idx is None, "lesson_idx must be None if curriculum is None"
			return self.plot_history(TrainingHistoriesMap.REPORT_KEY, save_path, show, **kwargs)
		if lesson_idx is None:
			self.plot_history(TrainingHistoriesMap.REPORT_KEY, save_path, show, **kwargs)
		else:
			self.plot_history(self.curriculum[lesson_idx].name, save_path, show, **kwargs)

	def plot_history(
			self,
			history_name: str,
			save_path=None,
			show=False,
			**kwargs
	):
		os.makedirs(os.path.dirname(save_path), exist_ok=True)
		history = self.histories[history_name]
		if self.curriculum is not None and history_name != TrainingHistoriesMap.REPORT_KEY:
			lessons = [self.curriculum[history_name]]
			lessons_start_itr = [0]
		elif self.curriculum is not None and history_name == TrainingHistoriesMap.REPORT_KEY:
			lessons = self.curriculum.lessons
			lessons_lengths = {k: [len(self.histories[lesson.name][k]) for lesson in lessons] for k in history._container}
			lessons_start_itr = {k: np.cumsum(lessons_lengths[k]) for k in history.keys()}
		else:
			lessons = []
			lessons_start_itr = []

		kwargs = self._set_default_plot_kwargs(kwargs)
		loss_metrics = [k for k in history.keys() if 'loss' in k.lower()]
		rewards_metrics = [k for k in history.keys() if 'reward' in k.lower()]
		other_metrics = [k for k in history.keys() if k not in loss_metrics and k not in rewards_metrics]
		n_metrics = 2 + len(other_metrics)
		n_cols = int(np.sqrt(n_metrics))
		n_rows = int(n_metrics / n_cols)
		fig, axes = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=kwargs["figsize"], sharex='all')
		if axes.ndim == 1:
			axes = np.expand_dims(axes, axis=-1)
		for row_i in range(n_rows):
			for col_i in range(n_cols):
				ax = axes[row_i, col_i]
				ravel_index = row_i * n_cols + col_i
				if ravel_index == 0:
					for k in loss_metrics:
						ax.plot(history[k], label=k, linewidth=kwargs['linewidth'])
					ax.set_ylabel("Loss [-]", fontsize=kwargs["fontsize"])
					ax.legend(fontsize=kwargs["fontsize"])
				elif ravel_index == 1:
					for k in rewards_metrics:
						ax.plot(history[k], label=k, linewidth=kwargs['linewidth'])
						for lesson_idx, lesson in enumerate(lessons):
							if lesson.completion_criteria.measure == k:
								ax.plot(
									lesson.completion_criteria.threshold*np.ones(len(history[k])), 'k--',
									label=f"{k} threshold", linewidth=kwargs['linewidth']
								)
							if history_name == TrainingHistoriesMap.REPORT_KEY and lesson.is_completed:
								ax.axvline(
									lessons_start_itr[k][lesson_idx], ymin=np.min(history[k]), ymax=np.max(history[k]),
									color='r', linestyle='--', linewidth=kwargs['linewidth'], label=f"lesson start"
								)
					ax.set_ylabel("Rewards [-]", fontsize=kwargs["fontsize"])
					ax.legend(fontsize=kwargs["fontsize"])
				else:
					k = other_metrics[ravel_index - 1]
					ax.plot(history[k], label=k, linewidth=kwargs['linewidth'])
					ax.legend(fontsize=kwargs["fontsize"])
				if row_i == n_rows - 1:
					ax.set_xlabel("Iterations [-]", fontsize=kwargs["fontsize"])
				legend_without_duplicate_labels_(ax)
		if save_path is not None:
			plt.savefig(save_path, dpi=kwargs["dpi"])
		if show:
			plt.show()
		plt.close(fig)







from collections import defaultdict
from typing import Optional, Sequence, Union, Dict, Callable, List, Tuple

import numpy as np
import torch
import torch.nn.functional as F

from .learning_algorithm import LearningAlgorithm
from ..learning_algorithms.tbptt import TBPTT
from ..transforms.base import ToDevice
from ..utils import compute_jacobian, list_insert_replace_at


class RLS(TBPTT):
	r"""
	Apply the recursive least squares algorithm to the given model. Different strategies are available to update the
	parameters of the model. The strategy is defined by the :attr:`strategy` attribute of the class. The following
	strategies are available:
		
		- `inputs`: The parameters are updated using the inputs of the model.
		- `outputs`: The parameters are updated using the outputs of the model. This one is inspired by the work of
			Perich and al. :cite:t:`perich_inferring_2021` with the CURBD algorithm.
		- `grad`: The parameters are updated using the gradients of the model. This one is inspired by the work of
			Zhang and al. :cite:t:`zhang_revisiting_2021`.
		- `jacobian`: The parameters are updated using the Jacobian of the model. This one is inspired by the work of
			Al-Batah and al. :cite:t:`al-batah_modified_2010`.
		- `scaled_jacobian`: The parameters are updated using the scaled Jacobian of the model.
	
	.. note::
		The `inputs` and `outputs` strategies are limited to an optimization of only one parameter. The others
		strategies can be used with multiple parameters. Unfortunately, those strategies do not work as expected
		at the moment. If you want to help with the development of those strategies, please open an issue on
		GitHub.
	
	"""
	CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"
	CHECKPOINT_P_STATES_DICT_KEY: str = "P_list"
	
	def __init__(
			self,
			*,
			params: Optional[Sequence[torch.nn.Parameter]] = None,
			layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
			criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] = None,
			backward_time_steps: Optional[int] = None,
			is_recurrent: bool = False,
			**kwargs
	):
		"""
		Constructor for RLS class.

		:param params: The parameters to optimize. If None, the parameters of the model's trainer will be used.
		:type params: Optional[Sequence[torch.nn.Parameter]]
		:param layers: The layers to optimize. If not Noen the parameters of the layers wiill be added to the
			parameters to optimize.
		:type layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]]
		:param criterion: The criterion to use. If not provided, torch.nn.MSELoss is used.
		:type criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]]
		:param backward_time_steps: The frequency of parameter optimisation. If None, the number of
			time steps of the data will be used.
		:type backward_time_steps: Optional[int]
		:param is_recurrent: If True, the model is recurrent. If False, the model is not recurrent.
		:type is_recurrent: bool
		:param kwargs: The keyword arguments to pass to the BaseCallback.

		:keyword bool save_state: Whether to save the state of the optimizer. Defaults to True.
		:keyword bool load_state: Whether to load the state of the optimizer. Defaults to True.
		"""
		kwargs.setdefault("auto_backward_time_steps_ratio", 0)
		kwargs.setdefault("save_state", True)
		kwargs.setdefault("load_state", True)
		super().__init__(
			params=params,
			layers=layers,
			criterion=criterion,
			backward_time_steps=backward_time_steps,
			optimizer=None,
			optim_time_steps=None,
			**kwargs
		)
		
		# RLS attributes
		self.P_list = None
		self.delta = kwargs.get("delta", 1.0)
		self.Lambda = kwargs.get("Lambda", 1.0)
		self.kappa = kwargs.get("kappa", 1.0)
		self._device = kwargs.get("device", None)
		self.to_cpu_transform = ToDevice(device=torch.device("cpu"))
		self.to_device_transform = None
		self._other_dims_as_batch = kwargs.get("other_dims_as_batch", False)
		self._is_recurrent = is_recurrent
		self.strategy = kwargs.get("strategy", "inputs").lower()
		self.strategy_to_mth = {
			"inputs": self.inputs_mth_step,
			"outputs": self.outputs_mth_step,
			"grad": self.grad_mth_step,
			"jacobian": self.jacobian_mth_step,
			"scaled_jacobian": self.scaled_jacobian_mth_step,
		}
		self.kwargs = kwargs
		self._asserts()
		self._last_layers_buffer = defaultdict(list)
	
	def _asserts(self):
		assert 0.0 < self.Lambda <= 1.0, "Lambda must be between 0 and 1"
		assert self.strategy in self.strategy_to_mth, f"Strategy must be one of {list(self.strategy_to_mth.keys())}"
		
	def __repr__(self):
		repr_str = f"{self.name}: ("
		repr_str += f"priority={self.priority}, "
		repr_str += f"save_state={self.save_state}, "
		repr_str += f"load_state={self.load_state}, "
		repr_str += f"strategy={self.strategy}, "
		repr_str += f"delta={self.delta}, "
		repr_str += f"Lambda={self.Lambda}, "
		repr_str += f"kappa={self.kappa}"
		repr_str += f")"
		return repr_str
	
	def initialize_P_list(self, m=None):
		self.P_list = [
			self.delta * torch.eye(param.numel() if m is None else m, dtype=torch.float32, device=torch.device("cpu"))
			for param in self.params
		]
	
	def _maybe_update_time_steps(self):
		if self._auto_set_backward_time_steps:
			self.backward_time_steps = max(1, int(self._auto_backward_time_steps_ratio * self._data_n_time_steps))
			
	def _decorate_forward(self, forward, layer_name: str):
		def _forward(*args, **kwargs):
			out = forward(*args, **kwargs)
			t = kwargs.get("t", None)
			if t is None:
				return out
			out_tensor = self._get_out_tensor(out)
			if t == 0:  # Hotfix for the first time step  TODO: fix this
				ready = bool(self._layers_buffer[layer_name])
			else:
				ready = True
			list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
			if len(self._layers_buffer[layer_name]) == self.backward_time_steps and ready:
				self._backward_at_t(t, self.backward_time_steps, layer_name)
				if self.strategy in ["grad", "jacobian", "scaled_jacobian"]:
					out = self._detach_out(out)
			return out
		return _forward
	
	def _backward_at_t(self, t: int, backward_t: int, layer_name: str):
		if self._last_layers_buffer[layer_name]:
			x_batch = self._get_x_batch_from_buffer(layer_name)
		else:
			x_batch = self._get_x_batch_slice_from_trainer(0, backward_t, layer_name)
		y_batch = self._get_y_batch_slice_from_trainer((t + 1) - backward_t, t + 1, layer_name)
		pred_batch = self._get_pred_batch_from_buffer(layer_name)
		self.optimization_step(x_batch, pred_batch, y_batch)
		self._last_layers_buffer[layer_name] = self._layers_buffer[layer_name].copy()
		self._layers_buffer[layer_name].clear()
	
	def _get_x_batch_slice_from_trainer(self, t_first: int, t_last: int, layer_name: str = None):
		x_batch = self.trainer.current_training_state.x_batch
		if isinstance(x_batch, dict):
			if layer_name is None:
				x_batch = {
					key: val[:, t_first:t_last]
					for key, val in x_batch.items()
				}
			else:
				x_batch = x_batch[layer_name][:, t_first:t_last]
		else:
			x_batch = x_batch[:, t_first:t_last]
		return x_batch.clone()
	
	def _get_x_batch_from_buffer(self, layer_name: str):
		pred_batch = torch.stack(self._last_layers_buffer[layer_name], dim=1)
		return pred_batch
	
	def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
		if self.save_state:
			state = checkpoint.get(self.name, {})
			self.P_list = state.get(self.CHECKPOINT_P_STATES_DICT_KEY, None)
			
	def get_checkpoint_state(self, trainer, **kwargs) -> object:
		if self.save_state:
			return {
				self.CHECKPOINT_P_STATES_DICT_KEY: self.P_list,
			}
		return None
	
	def _try_put_on_device(self, trainer):
		try:
			self.P_list = [self.to_device_transform(p) for p in self.P_list]
		except Exception as e:
			trainer.model = self.to_cpu_transform(trainer.model)
			self.P_list = [self.to_device_transform(p) for p in self.P_list]
	
	def _put_on_cpu(self):
		self.P_list = [self.to_cpu_transform(p) for p in self.P_list]
	
	def start(self, trainer, **kwargs):
		LearningAlgorithm.start(self, trainer, **kwargs)
		if self.params and self.optimizer is None:
			self.optimizer = torch.optim.SGD(self.params, lr=self.kwargs.get("lr", 1.0))
		elif not self.params and self.optimizer is not None:
			self.params.extend(
				[
					param
					for i in range(len(self.optimizer.param_groups))
					for param in self.optimizer.param_groups[i]["params"]
				]
			)
		else:
			self.params = list(trainer.model.parameters())
			self.optimizer = torch.optim.SGD(self.params, lr=self.kwargs.get("lr", 1.0))
		
		if self.criterion is None and trainer.criterion is not None:
			self.criterion = trainer.criterion
			
		# filter params to get only the ones that require gradients
		self.params = [param for param in self.params if param.requires_grad]
		
		if self._device is None:
			self._device = trainer.model.device
		self.to_device_transform = ToDevice(device=self._device)
		self.output_layers: torch.nn.ModuleDict = trainer.model.output_layers
		self._initialize_original_forwards()
	
	def on_batch_begin(self, trainer, **kwargs):
		LearningAlgorithm.on_batch_begin(self, trainer, **kwargs)
		self.trainer = trainer
		if self._is_recurrent:
			self._data_n_time_steps = self._get_data_time_steps_from_y_batch(trainer.current_training_state.y_batch)
			self._maybe_update_time_steps()
			self.decorate_forwards()
	
	def on_batch_end(self, trainer, **kwargs):
		super().on_batch_end(trainer)
		self.undecorate_forwards()
		self._layers_buffer.clear()
	
	def on_optimization_begin(self, trainer, **kwargs):
		x_batch = trainer.current_training_state.x_batch
		y_batch = trainer.current_training_state.y_batch
		pred_batch = trainer.format_pred_batch(trainer.current_training_state.pred_batch, y_batch)
		
		if self._is_recurrent:
			for layer_name in self._layers_buffer:
				backward_t = len(self._layers_buffer[layer_name])
				if backward_t > 0:
					self._backward_at_t(self._data_n_time_steps - 1, backward_t, layer_name)
		else:
			self.optimization_step(x_batch, pred_batch, y_batch)
		
		trainer.update_state_(batch_loss=self.apply_criterion(pred_batch, y_batch).detach_())
	
	def optimization_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
		if self.strategy not in self.strategy_to_mth:
			raise ValueError(f"Invalid strategy: {self.strategy}")
		return self.strategy_to_mth[self.strategy](x_batch, pred_batch, y_batch)
	
	def scaled_jacobian_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
		"""
		This method is inspired by the work of Al-Batah and al. :cite:t:`al-batah_modified_2010`. Unfortunately, this
		method does not seem to work with the current implementation.

		TODO: Make it work.

		x.shape = [B, f_in]
		y.shape = [B, f_out]
		error.shape = [B, f_out]
		P.shape = [f_out, f_out]
		theta.shape = [ell, 1]

		epsilon = mean[B](error[B, f_out]) -> [1, f_out]
		phi = mean[B](y[B, f_out]) [1, f_out]
		psi = jacobian[theta](phi[1, f_out]]) -> [f_out, ell]

		K = P[f_out, f_out] @ psi[f_out, ell] -> [f_out, ell]
		h = 1 / (labda[1] + kappa[1] * psi.T[ell, f_out] @ K[f_out, ell]) -> [ell, ell]
		grad = (K[f_out, ell] @ h[ell, ell]).T[ell, f_out] @ epsilon.T[f_out, 1] -> [1, ell]
		P = labda[1] * P[f_out, f_out] - kappa[1] * K[f_out, ell] @ h[ell, ell] @ K[f_out, ell].T -> [f_out, f_out]

		In this case f_in must be equal to N_in.

		:param x_batch: inputs of the layer
		:param pred_batch: outputs of the layer
		:param y_batch: targets of the layer

		"""
		model_device = self.trainer.model.device
		assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
		assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
		assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
		self.optimizer.zero_grad()
		
		x_batch_view = x_batch.view(-1, x_batch.shape[-1])  # [B, f_in]
		pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1])  # [B, f_out]
		y_batch_view = y_batch.view(-1, y_batch.shape[-1])  # [B, f_out]
		error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]
		
		if self.P_list is None:
			self.initialize_P_list(m=pred_batch_view.shape[-1])
		self.P_list = self.to_device_transform(self.P_list)
		self.params = self.to_device_transform(self.params)
		
		epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
		phi = pred_batch_view.mean(dim=0).view(1, -1)  # [1, f_out]
		psi_list = compute_jacobian(params=self.params, y=phi.view(-1), strategy="slow")  # [f_out, ell]
		K_list = [torch.matmul(P, psi) for P, psi in zip(self.P_list, psi_list)]  # [f_out, f_out] @ [f_out, ell] -> [f_out, ell]
		h_list = [torch.linalg.pinv(self.Lambda + self.kappa * torch.matmul(psi.T, K)) for psi, K in zip(psi_list, K_list)]  # [ell, f_out] @ [f_out, ell] -> [ell, ell]
		
		for p, K, h in zip(self.params, K_list, h_list):
			p.grad = torch.matmul(torch.matmul(K, h).T, epsilon.T).view(p.shape).clone()  # ([f_out, ell] @ [ell, ell]).T @ [f_out, 1] -> [ell, 1]
		
		self.optimizer.step()
		self.P_list = [
			self.Lambda * P - self.kappa * torch.matmul(torch.matmul(K, h), K.T)
			for P, h, K in zip(self.P_list, h_list, K_list)
		]  # [f_out, f_out] - [f_out, ell] @ [ell, ell] @ [ell, f_out] -> [f_out, f_out]
		
		self._put_on_cpu()
		self.trainer.model.to(model_device, non_blocking=True)
	
	def jacobian_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
		"""
		This method is inspired by the work of Al-Batah and al. :cite:t:`al-batah_modified_2010`. Unfortunately, this
		method does not seem to work with the current implementation.

		TODO: Make it work.

		x.shape = [B, f_in]
		y.shape = [B, f_out]
		error.shape = [B, f_out]
		P.shape = [f_out, f_out]
		theta.shape = [ell, 1]

		epsilon = mean[B](error[B, f_out]) -> [1, f_out]
		phi = mean[B](y[B, f_out]) [1, f_out]
		psi = jacobian[theta](phi[1, f_out]]) -> [f_out, L]

		K = P[f_out, f_out] @ psi[f_out, L] -> [f_out, L]
		grad = epsilon[1, f_out] @ K[f_out, L] -> [L, 1]
		P = labda[1] * P[f_out, f_out] - kappa[1] * K[f_out, L] @ K[f_out, L].T -> [f_out, f_out]

		In this case f_in must be equal to N_in.

		:param x_batch: inputs of the layer
		:param pred_batch: outputs of the layer
		:param y_batch: targets of the layer

		"""
		model_device = self.trainer.model.device
		assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
		assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
		assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
		self.optimizer.zero_grad()
		
		x_batch_view = x_batch.view(-1, x_batch.shape[-1])  # [B, f_in]
		pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1])  # [B, f_out]
		y_batch_view = y_batch.view(-1, y_batch.shape[-1])  # [B, f_out]
		error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]
		
		if self.P_list is None:
			self.initialize_P_list(m=pred_batch_view.shape[-1])
		self.P_list = self.to_device_transform(self.P_list)
		self.params = self.to_device_transform(self.params)
		
		epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
		phi = pred_batch_view.mean(dim=0).view(1, -1)  # [1, f_out]
		psi_list = compute_jacobian(params=self.params, y=phi.view(-1), strategy="slow")  # [f_out, L]
		K_list = [torch.matmul(P, psi) for P, psi in zip(self.P_list, psi_list)]  # [f_out, f_out] @ [f_out, ell] -> [f_out, L]
		
		for p, K in zip(self.params, K_list):
			p.grad = torch.matmul(K.T, epsilon.T).view(p.shape).clone()  # [L, f_out] @ [f_out, 1] -> [L, 1]
		
		self.optimizer.step()
		self.P_list = [
			self.Lambda * P - self.kappa * torch.matmul(K, K.T)
			for P, K in zip(self.P_list, K_list)
		]  # [f_out, f_out] - [f_out, L] @ [L, f_out] -> [f_out, f_out]
		
		self._put_on_cpu()
		self.trainer.model.to(model_device, non_blocking=True)
	
	def grad_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
		"""
		This method is inspired by the work of Zhang and al. :cite:t:`zhang_revisiting_2021`. Unfortunately, this
		method does not seem to work with the current implementation.
		
		TODO: Make it work.

		x.shape = [B, f_in]
		y.shape = [B, f_out]
		error.shape = [B, f_out]
		P.shape = [f_in, f_in]
		
		epsilon = mean[B](error[B, f_out]) -> [1, f_out]
		phi = mean[B](x[B, f_in]) [1, f_in]

		K = P[f_in, f_in] @ phi.T[f_in, 1] -> [f_in, 1]
		h = 1 / (labda[1] + kappa[1] * phi[1, f_in] @ K[f_in, 1]) -> [1]
		grad = h[1] * P[f_in, f_in] @ grad[N_in, N_out] -> [N_in, N_out]
		P = labda[1] * P[f_in, f_in] - h[1] * kappa[1] * K[f_in, 1] @ K.T[1, f_in] -> [f_in, f_in]

		In this case f_in must be equal to N_in.

		:param x_batch: inputs of the layer
		:param pred_batch: outputs of the layer
		:param y_batch: targets of the layer

		"""
		model_device = self.trainer.model.device
		assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
		assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
		assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
		self.optimizer.zero_grad()
		
		mse_loss = F.mse_loss(pred_batch, y_batch)
		mse_loss.backward()
		
		x_batch_view = x_batch.view(-1, x_batch.shape[-1])  # [B, f_in]
		pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1])  # [B, f_out]
		y_batch_view = y_batch.view(-1, y_batch.shape[-1])  # [B, f_out]
		error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]
		
		if self.P_list is None:
			self.initialize_P_list(m=x_batch_view.shape[-1])
			for p in self.params:
				# making sur that f_in = N_in.
				if p.shape[0] != x_batch_view.shape[-1]:
					raise ValueError(
						f"For inputs of shape [B, f_in], the first dimension of the parameters must be f_in, "
						f"got {p.shape[0]} instead of {x_batch_view.shape[-1]}."
					)
		self.P_list = self.to_device_transform(self.P_list)
		self.params = self.to_device_transform(self.params)
		
		epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
		phi = x_batch_view.mean(dim=0).view(1, -1).detach().clone()  # [1, f_in]
		K_list = [torch.matmul(P, phi.T) for P in self.P_list]  # [f_in, f_in] @ [f_in, 1] -> [f_in, 1]
		h_list = [1.0 / (self.Lambda + self.kappa * torch.matmul(phi, K)).item() for K in K_list]  # [1, f_in] @ [f_in, 1] -> [1]
		
		for p, P, h in zip(self.params, self.P_list, h_list):
			p.grad = h * torch.matmul(P, p.grad)  # [f_in, f_in] @ [N_in, N_out] -> [N_in, N_out]
		
		self.optimizer.step()
		self.P_list = [
			self.Lambda * P - h * self.kappa * torch.matmul(K, K.T)
			for P, h, K in zip(self.P_list, h_list, K_list)
		]  # [f_in, 1] @ [1, f_in] -> [f_in, f_in]
		
		self._put_on_cpu()
		self.trainer.model.to(model_device, non_blocking=True)
	
	def inputs_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
		"""
		

		x.shape = [B, f_in]
		y.shape = [B, f_out]
		error.shape = [B, f_out]
		P.shape = [f_in, f_in]
		
		epsilon = mean[B](error[B, f_out]) -> [1, f_out]
		phi = mean[B](x[B, f_in]) [1, f_in]

		K = P[f_in, f_in] @ phi.T[f_in, 1] -> [f_in, 1]
		h = 1 / (labda[1] + kappa[1] * phi[1, f_in] @ K[f_in, 1]) -> [1]
		P = labda[1] * P[f_in, f_in] - h[1] * kappa[1] * K[f_in, 1] @ K.T[1, f_in] -> [f_in, f_in]
		grad = h[1] * K[f_in, 1] @ epsilon[1, f_out] -> [N_in, N_out]
		
		In this case [N_in, N_out] must be equal to [f_in, f_out].

		:param x_batch: inputs of the layer
		:param pred_batch: outputs of the layer
		:param y_batch: targets of the layer

		"""
		model_device = self.trainer.model.device
		assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
		assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
		assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
		self.optimizer.zero_grad()
		
		x_batch_view = x_batch.view(-1, x_batch.shape[-1])  # [B, f_in]
		pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1])  # [B, f_out]
		y_batch_view = y_batch.view(-1, y_batch.shape[-1])  # [B, f_out]
		error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]
		
		if self.P_list is None:
			self.initialize_P_list(m=x_batch_view.shape[-1])
			for p in self.params:
				# making sur that f_in = N_in.
				if p.shape[0] != x_batch_view.shape[-1]:
					raise ValueError(
						f"For inputs of shape [B, f_in], the first dimension of the parameters must be f_in, "
						f"got {p.shape[0]} instead of {x_batch_view.shape[-1]}."
					)
				# making sure that f_out = N_out.
				if p.shape[1] != y_batch_view.shape[-1]:
					raise ValueError(
						f"For targets of shape [B, f_out], the second dimension of the parameters must be f_out, "
						f"got {p.shape[1]} instead of {y_batch_view.shape[-1]}."
					)
		self.P_list = self.to_device_transform(self.P_list)
		self.params = self.to_device_transform(self.params)
		
		epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
		phi = x_batch_view.mean(dim=0).view(1, -1).detach().clone()  # [1, f_in]
		K_list = [torch.matmul(P, phi.T) for P in self.P_list]  # [f_in, f_in] @ [f_in, 1] -> [f_in, 1]
		h_list = [1.0 / (self.Lambda + self.kappa * torch.matmul(phi, K)).item() for K in K_list]  # [1, f_in] @ [f_in, 1] -> [1]
		
		for p, K, h in zip(self.params, K_list, h_list):
			p.grad = h * torch.outer(K.view(-1), epsilon.view(-1))  # [f_in, 1] @ [1, f_out] -> [N_in, N_out]

		self.optimizer.step()
		self.P_list = [
			self.Lambda * P - h * self.kappa * torch.matmul(K, K.T)
			for P, h, K in zip(self.P_list, h_list, K_list)
		]  # [f_in, 1] @ [1, f_in] -> [f_in, f_in]
		
		self._put_on_cpu()
		self.trainer.model.to(model_device, non_blocking=True)
	
	def outputs_mth_step(self, x_batch: torch.Tensor, pred_batch: torch.Tensor, y_batch: torch.Tensor):
		"""
		This method is inspired by the work of Perich and al. :cite:t:`perich_inferring_2021` with
		the CURBD algorithm.

		x.shape = [B, f_in]
		y.shape = [B, f_out]
		error.shape = [B, f_out]
		epsilon = mean[B](error[B, f_out]) -> [1, f_out]
		phi = mean[B](y[B, f_out]) [1, f_out]

		P.shape = [f_out, f_out]
		K = P[f_out, f_out] @ phi.T[f_out, 1] -> [f_out, 1]
		h = 1 / (labda[1] + kappa[1] * phi[1, f_out] @ K[f_out, 1]) -> [1]
		P = labda[1] * P[f_out, f_out] - h[1] * kappa[1] * K[f_out, 1] @ K.T[1, f_out] -> [f_out, f_out]
		grad = h[1] * K[f_out, 1] @ epsilon[1, f_out] -> [N_in, N_out]
		
		In this case [N_in, N_out] must be equal to [f_out, f_out].

		:param x_batch: inputs of the layer
		:param pred_batch: outputs of the layer
		:param y_batch: targets of the layer

		"""
		model_device = self.trainer.model.device
		assert isinstance(x_batch, torch.Tensor), "x_batch must be a torch.Tensor"
		assert isinstance(pred_batch, torch.Tensor), "pred_batch must be a torch.Tensor"
		assert isinstance(y_batch, torch.Tensor), "y_batch must be a torch.Tensor"
		self.optimizer.zero_grad()
		
		x_batch_view = x_batch.view(-1, x_batch.shape[-1])  # [B, f_in]
		pred_batch_view = pred_batch.view(-1, pred_batch.shape[-1])  # [B, f_out]
		y_batch_view = y_batch.view(-1, y_batch.shape[-1])  # [B, f_out]
		error = self.to_device_transform(pred_batch_view - y_batch_view)  # [B, f_out]
		
		if self.P_list is None:
			self.initialize_P_list(m=pred_batch_view.shape[-1])
			for p in self.params:
				# making sur that f_out = N_in.
				if p.shape[0] != pred_batch_view.shape[-1]:
					raise ValueError(
						f"For inputs of shape [B, f_in], the first dimension of the parameters must be f_in, "
						f"got {p.shape[0]} instead of {x_batch_view.shape[-1]}."
					)
				# making sure that f_out = N_out.
				if p.shape[1] != pred_batch_view.shape[-1]:
					raise ValueError(
						f"For targets of shape [B, f_out], the second dimension of the parameters must be f_out, "
						f"got {p.shape[1]} instead of {y_batch_view.shape[-1]}."
					)
		self.P_list = self.to_device_transform(self.P_list)
		self.params = self.to_device_transform(self.params)
		
		epsilon = error.mean(dim=0).view(1, -1)  # [1, f_out]
		phi = pred_batch_view.mean(dim=0).view(1, -1).detach().clone()  # [1, f_out]
		K_list = [torch.matmul(P, phi.T) for P in self.P_list]  # [f_out, f_out] @ [f_out, 1] -> [f_out, 1]
		h_list = [1.0 / (self.Lambda + self.kappa * torch.matmul(phi, K)).item() for K in K_list]  # [1, f_out] @ [f_out, 1] -> [1]
		
		for p, K, h in zip(self.params, K_list, h_list):
			p.grad = h * torch.outer(K.view(-1), epsilon.view(-1))  # [f_out, 1] @ [1, f_out] -> [N_in, N_out]
		
		self.optimizer.step()
		self.P_list = [
			self.Lambda * P - h * self.kappa * torch.matmul(K, K.T)
			for P, h, K in zip(self.P_list, h_list, K_list)
		]  # [f_in, 1] @ [1, f_in] -> [f_in, f_in]
		
		self._put_on_cpu()
		self.trainer.model.to(model_device, non_blocking=True)




from typing import Any, Type, Union, Optional

import numpy as np
import torch

from . import ConstantValuesTransform, to_tensor
from ..modules.layers import LIFLayer, SpyLIFLayer, ALIFLayer


class SpikesEncoder(torch.nn.Module):
	def __init__(
			self,
			n_steps: int,
			n_units: int,
			spikes_layer_type: Type[Union[LIFLayer, SpyLIFLayer, ALIFLayer]],
			dt: float = 1e-3,
			device: Optional[torch.device] = None,
			**kwargs
	):
		"""
		Constructor for the SpikesEncoder class.
		
		:param n_steps: Number of steps to encode.
		:param n_units: Number of units in the encoder.
		:param spikes_layer_type: Dynamic of the spikes layer.
		:param dt: Time step of the simulation.
		:param device: Device to use for the encoder.
		:param kwargs: Keyword arguments for the spikes layer.
		
		:keyword kwargs:
			* <forward_weights>: Forward weights of the spikes layer. Default: eye(n_units).
			* <use_recurrent_connection>: Whether to use a recurrent connection. Default: False.
			* <name>: Name of the spikes layer. Default: "encoder".
		
		"""
		super().__init__()
		self.n_steps = n_steps
		self.const_transform = ConstantValuesTransform(n_steps, batch_wise=True)
		self.spikes_layer_type = spikes_layer_type
		self.dt = dt
		kwargs.setdefault('forward_weights', torch.eye(n_units))
		kwargs.setdefault('use_recurrent_connection', False)
		kwargs.setdefault('name', 'encoder')
		kwargs.setdefault('freeze_weights', False)
		
		assert "dt" not in kwargs, \
			"dt cannot be specified since it must be the given dt"
		assert "device" not in kwargs, \
			"device cannot be specified since it must be the given device"
		
		self.spikes_layer = self.spikes_layer_type(
			n_units, n_units,
			dt=dt,
			device=device,
			**kwargs,
		)
		self.spikes_layer.build()
	
	def to(self, device):
		self.spikes_layer.device = device
		return super().to(device)

	def forward(self, x: Any):
		x = to_tensor(x).to(self.spikes_layer.device)
		x = self.const_transform(x)
		x_spikes, hh = [], None
		for t in range(self.n_steps):
			spikes, hh = self.spikes_layer(x[:, t], hh)
			x_spikes.append(spikes)
		# x_spikes = torch.squeeze(torch.stack(x_spikes, dim=1))
		return torch.stack(x_spikes, dim=1)
	
	def get_and_reset_regularization_loss(self) -> torch.Tensor:
		"""
		Get the regularization loss as a sum of all the regularization losses of the layers. Then reset the
		regularization losses.

		:return: the regularization loss.
		"""
		regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.spikes_layer.device)
		if hasattr(self.spikes_layer, "get_and_reset_regularization_loss") and callable(
				self.spikes_layer.get_and_reset_regularization_loss
			):
			regularization_loss += self.spikes_layer.get_and_reset_regularization_loss()
		return regularization_loss


class LIFEncoder(SpikesEncoder):
	def __init__(
			self,
			n_steps: int,
			n_units: int,
			dt: float = 1e-3,
			device: Optional[torch.device] = None,
			spikes_layer_kwargs: Optional[dict] = None,
	):
		if spikes_layer_kwargs is None:
			spikes_layer_kwargs = {}
		super().__init__(
			n_steps, n_units, LIFLayer, dt, device, **spikes_layer_kwargs,
		)


class SpyLIFEncoder(SpikesEncoder):
	def __init__(
			self,
			n_steps: int,
			n_units: int,
			dt: float = 1e-3,
			device: Optional[torch.device] = None,
			spikes_layer_kwargs: Optional[dict] = None,
	):
		if spikes_layer_kwargs is None:
			spikes_layer_kwargs = {}
		super().__init__(
			n_steps, n_units, SpyLIFLayer, dt, device, **spikes_layer_kwargs,
		)


class ALIFEncoder(SpikesEncoder):
	def __init__(
			self,
			n_steps: int,
			n_units: int,
			dt: float = 1e-3,
			device: Optional[torch.device] = None,
			spikes_layer_kwargs: Optional[dict] = None,
	):
		if spikes_layer_kwargs is None:
			spikes_layer_kwargs = {}
		super().__init__(
			n_steps, n_units, ALIFLayer, dt, device, **spikes_layer_kwargs,
		)





import torch
import torch.nn as nn
from typing import Optional, Union, Iterable, Dict
import pythonbasictools as pybt
from . import BaseRegularization
from ..transforms.base import to_numpy
from ..init import dale_


# @pybt.docstring.inherit_fields_docstring(fields=["Attributes"], bases=[BaseRegularization])
class DaleLawL2(BaseRegularization):
	"""
	Regularisation of the connectome to apply Dale's law and L2. In a nutshell, the Dale's law
	stipulate that neurons can either have excitatory or inhibitory connections, not both.
	The L2 regularisation reduce the energy of the network. This regularisation allow you to follow the
	Dale's law and/or L2 depending on the factor alpha. The equation used is showed by :eq:`dale_l2`.
	
	.. math::
		:label: dale_l2
		
		\\begin{equation}
			\\mathcal{L}_{\\text{DaleLawL2}} = \\text{Tr}\\left( W^T \\left(\\alpha W - \\left(1 - \\alpha\\right) W_{\\text{ref}}\\right) \\right)
		\\end{equation}
	
	
	In the case where :math:`\\alpha = 0`, the regularisation will only follow the Dale's law shown by :eq:`dale`.
	
	.. math::
		:label: dale
		
		\\begin{equation}
			\\mathcal{L}_{\\text{DaleLaw}} = -\\text{Tr}\\left( W^T W_{\\text{ref}}\\right)
		\\end{equation}
	
	In the case where :math:`\\alpha = 1`, the regularisation will only follow the L2 regularisation shown by :eq:`l2`.
	
	.. math::
		:label: l2
		
		\\begin{equation}
			\\mathcal{L}_{\\text{L2}} = \\text{Tr}\\left( W^T W\\right)
		\\end{equation}
	
	
	:Attributes:
		- :attr:`alpha` (float): Number between 0 and 1 that favors one of the constraints.
		- :attr:`dale_kwargs` (dict): kwargs of the Dale's law. See :func:`dale_`.
		- :attr:`reference_weights` (Iterable[torch.Tensor]): Reference weights to compare. Must be the same size as the weights.
	
	"""
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			alpha: float = 0.8,
			reference_weights: Optional[Iterable[torch.Tensor]] = None,
			Lambda: float = 1.0,
			optimizer: Optional[torch.optim.Optimizer] = None,
			**dale_kwargs
	):
		"""
		:param params: Weights matrix to regularize (can be multiple)
		:type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
		:param alpha: Number between 0 and 1 that favors one of the constraints.
			If alpha = 0 -> Only Dale's law is applied.
			If alpha = 1 -> Only the reduction of the energy is applied.
			If 1 < alpha < 0 -> Both Dale's law and the reduction of the energy are applied with their ratio.
		:type alpha: float
		:param reference_weights: Reference weights to compare. Must be the same size as the weights. If not provided,
			the weights will be generated automatically with the dale_kwargs.
		:type reference_weights: Optional[Iterable[torch.Tensor]]
		:param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
		:type Lambda: float
		:param dale_kwargs: kwargs of the Dale's law.
		
		:keyword float inh_ratio: ratio of inhibitory connections. Must be between 0 and 1.
		:keyword float rho: The connectivity ratio. Must be between 0 and 1. If rho = 1, the tensor will be fully connected.
		:keyword bool inh_first: If True, the inhibitory neurons will be in the first half of the tensor. If False,
			the neurons will be shuffled.
		:keyword Optional[int] seed: seed for the random number generator. If None, the seed is not set.
		"""
		super(DaleLawL2, self).__init__(params, Lambda, optimizer=optimizer)
		self.__name__ = self.__class__.__name__
		self.alpha = alpha
		if self.alpha > 1 or self.alpha < 0:
			raise ValueError("alpha must be between 0 and 1")
		self.dale_kwargs = dale_kwargs
		self.reference_weights = self._init_reference_weights(reference_weights)
	
	def _init_reference_weights(
			self,
			reference_weights: Optional[Union[Iterable[torch.Tensor], Dict[str, torch.Tensor]]] = None
	):
		"""
		Initialize the reference weights with Dale's law.
		"""
		if reference_weights is None:
			self.reference_weights = []
			for param in self.params:
				self.reference_weights.append(torch.sign(dale_(torch.empty_like(param), **self.dale_kwargs)))
		else:
			self.reference_weights = [torch.sign(ref) for ref in reference_weights]
		return self.reference_weights

	def forward(self, *args, **kwargs) -> torch.Tensor:
		"""
		Compute the forward pass of the Dale's law. If alpha = 1 and the reference weights is not provided, it will be
		modified to 0, so it can get cancel.
		
		:param args: weights matrix
		:param kwargs: kwargs of the forward pass
		"""
		loss_list = []
		for param, ref in zip(self.params, self.reference_weights):
			loss = torch.trace(
				param.T @ (self.alpha * param - (1 - self.alpha) * ref.to(param.device))
			)
			loss_list.append(loss)
		if len(self.params) == 0:
			loss = torch.tensor(0.0, dtype=torch.float32)
		else:
			loss = torch.sum(torch.stack(loss_list))
		return loss


class DaleLaw(DaleLawL2):
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			reference_weights: Optional[Iterable[torch.Tensor]] = None,
			Lambda: float = 1.0,
			**dale_kwargs
	):
		"""
		:param params: Weights matrix to regularize (can be multiple)
		:param reference_weights: Reference weights to compare. Must be the same size as the weights. If not provided,
			the weights will be generated automatically with the dale_kwargs.
		:param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
		:param dale_kwargs: kwargs of the Dale's law.

		:keyword inh_ratio: ratio of inhibitory connections. Must be between 0 and 1.
		:keyword rho: The connectivity ratio. Must be between 0 and 1. If rho = 1, the tensor will be fully connected.
		:keyword inh_first: If True, the inhibitory neurons will be in the first half of the tensor. If False,
			the neurons will be shuffled.
		:keyword seed: seed for the random number generator. If None, the seed is not set.
		"""
		super(DaleLaw, self).__init__(params, 0.0, reference_weights, Lambda, **dale_kwargs)
		self.__name__ = self.__class__.__name__


class WeightsDistance(BaseRegularization):
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			reference_weights: Iterable[torch.Tensor],
			Lambda: float = 1.0,
			p: int = 1,
	):
		super(WeightsDistance, self).__init__(params, Lambda)
		self.reference_weights = list(reference_weights)
		self.p = p

	def forward(self, *args, **kwargs) -> torch.Tensor:
		loss_list = []
		for param, ref in zip(self.params, self.reference_weights):
			loss_list.append(torch.linalg.norm(torch.abs(param - ref), self.p))
		if len(self.params) == 0:
			loss = torch.tensor(0.0, dtype=torch.float32)
		else:
			loss = torch.sum(torch.stack(loss_list))
		return loss


class ExecRatioTargetRegularization(BaseRegularization):
	r"""
	Applies the function:

	.. math::
		\text{loss}(x) = \lambda \cdot \sum_{i=1}^N \left|(\text{mean}(\text{sign}(x_i)) + 1) - 2\cdot\text{target} \right|
	
	Where :math:`x` is the list of input parameters, :math:`N` is the number of parameter, :math:`\text{sign}(x_i)` is the
	sign of the element :math:`x_i`, :math:`\text{mean}(\text{sign}(x_i))` is the mean of the signs of the elements in the
	tensor, :math:`\text{target}` is the target value and :math:`\lambda` is the weight of the regularization.
	
	
	Examples::
		
		>>> layer = WilsonCowanLayer(10, 10, force_dale_law=True)
		>>> m = ExecRatioTargetRegularization(params=[layer.forward_sign], Lambda=1.0, target=0.5)
		>>> loss = m()
	"""
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			exec_target_ratio: float = 0.8,
			Lambda: float = 1.0,
			**kwargs
	):
		super(ExecRatioTargetRegularization, self).__init__(params, Lambda, **kwargs)
		assert 0 < exec_target_ratio < 1, "exec_target_ratio must be between 0 and 1"
		self.exec_target_ratio = exec_target_ratio
		self.sign_func = torch.nn.Softsign()
	
	def forward(self, *args, **kwargs) -> torch.Tensor:
		loss_list = []
		for param in self.params:
			param_ratio = torch.mean(self.sign_func(param)) + 1
			loss_list.append(torch.abs(param_ratio - 2 * self.exec_target_ratio))
		if len(self.params) == 0:
			loss = torch.tensor(0.0, dtype=torch.float32, requires_grad=True)
		else:
			loss = torch.sum(torch.stack(loss_list))
		return loss
	
	def on_pbar_update(self, trainer, **kwargs) -> dict:
		loss = to_numpy(self().item())
		exec_ratio = to_numpy(((torch.mean(self.sign_func(self.params[0])) + 1)/2).item())
		return {"exec_ratio": exec_ratio, "exec_ratio_loss": loss}
	

class InhRatioTargetRegularization(ExecRatioTargetRegularization):
	def __init__(
			self,
			params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
			inh_target_ratio: float = 0.2,
			Lambda: float = 1.0,
	):
		assert 0 < inh_target_ratio < 1, "inh_target_ratio must be between 0 and 1"
		super(InhRatioTargetRegularization, self).__init__(
			params=params,
			Lambda=Lambda,
			exec_target_ratio=1 - inh_target_ratio,
		)
	
	def on_pbar_update(self, trainer, **kwargs) -> dict:
		loss = to_numpy(self().item())
		inh_ratio = to_numpy(((1 - torch.mean(self.sign_func(self.params[0])))/2).item())
		return {"inh_ratio": inh_ratio, "inh_ratio_loss": loss}



import warnings
from collections import defaultdict
from typing import Optional, Sequence, Union, Dict, Callable, Tuple, List

import torch
from unstable import unstable

from .learning_algorithm import LearningAlgorithm
from ..transforms.base import to_numpy
from ..learning_algorithms.tbptt import TBPTT
from ..utils import batchwise_temporal_filter, list_insert_replace_at, zero_grad_params


@unstable
class Eprop(TBPTT):
	r"""
	Apply the eligibility trace forward propagation (e-prop) :cite:t:`bellec_solution_2020`
	algorithm to the given model.
	"""
	CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"
	CHECKPOINT_RN_FEEDBACK_WEIGHTS_KEY: str = "rn_feedback_weights"
	
	def __init__(
			self,
			*,
			params: Optional[Sequence[torch.nn.Parameter]] = None,
			layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
			**kwargs
	):
		"""
		Constructor for Eprop class.

		:param params: The parameters to optimize. If None, the parameters of the model's trainer will be used.
		:type params: Optional[Sequence[torch.nn.Parameter]]
		:param optimizer: The optimizer to use. If not provided, torch.optim.SGD is used.
		:type optimizer: Optional[torch.optim.Optimizer]
		:param criterion: The criterion to use. If not provided, torch.nn.MSELoss is used.
		:type criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]]
		:param kwargs: The keyword arguments to pass to the BaseCallback.

		:keyword bool save_state: Whether to save the state of the optimizer. Defaults to True.
		:keyword bool load_state: Whether to load the state of the optimizer. Defaults to True.
		"""
		warnings.warn("Eprop is still in beta and may not work as expected or act exactly as BPTT.", DeprecationWarning)
		kwargs.setdefault("save_state", True)
		kwargs.setdefault("load_state", True)
		# TODO: implement a optim step at each `optim_time_steps` steps.
		assert "backward_time_steps" not in kwargs, f"{self.__class__} does not support backward_time_steps."
		assert "optim_time_steps" not in kwargs, f"{self.__class__} does not support optim_time_steps."
		assert params is None, f"{self.__class__} does not support params yet."
		assert layers is not None, f"{self.__class__} requires layers."
		super().__init__(
			params=params,
			layers=layers,
			backward_time_steps=1,
			optim_time_steps=1,
			**kwargs
		)
		self.random_feedbacks = kwargs.get("random_feedbacks", True)
		if not self.random_feedbacks:
			raise NotImplementedError("Non-random feedbacks are not implemented yet.")
		self.feedback_weights = None
		self.rn_gen = torch.Generator()
		self.rn_gen.manual_seed(kwargs.get("seed", 0))
		self.running_grads = None
		self.eligibility_traces = defaultdict(list)
		self.learning_signals = defaultdict(list)
		self.layers_to_params = defaultdict(list)
	
	def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
		if self.save_state:
			state = checkpoint.get(self.name, {})
			opt_state_dict = state.get(self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY, None)
			if opt_state_dict is not None:
				self.optimizer.load_state_dict(opt_state_dict)
			if self.random_feedbacks:
				self.feedback_weights = state.get(self.CHECKPOINT_RN_FEEDBACK_WEIGHTS_KEY, None)
	
	def get_checkpoint_state(self, trainer, **kwargs) -> object:
		if self.save_state:
			state = {}
			if self.optimizer is not None:
				state[self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY] = self.optimizer.state_dict()
			if self.random_feedbacks:
				state[self.CHECKPOINT_RN_FEEDBACK_WEIGHTS_KEY] = self.feedback_weights
			return state
		return None
	
	def initialize_running_grads(self):
		self.running_grads = {
			layer.name: [
				torch.zeros_like(p)
				for p in self.layers_to_params[layer.name]
			]
			for layer in self.layers
		}
	
	def initialize_feedback_weights(self, pred_batch: Optional[torch.Tensor] = None):
		if pred_batch is None:
			pred_batch = self.trainer.current_training_state.pred_batch
		self.feedback_weights = {
			layer.name: [
				torch.randn((pred_batch.shape[-1], p.shape[-1]), generator=self.rn_gen)
				for p in self.layers_to_params[layer.name]
			]
			for layer in self.layers
		}
	
	def start(self, trainer, **kwargs):
		LearningAlgorithm.start(self, trainer, **kwargs)
		if self.params and self.optimizer is None:
			self.optimizer = torch.optim.SGD(self.params, lr=self.kwargs.get("lr", 1.0e-3))
		elif not self.params and self.optimizer is not None:
			self.params.extend(
				[
					param
					for i in range(len(self.optimizer.param_groups))
					for param in self.optimizer.param_groups[i]["params"]
				]
			)
		else:
			self.params = list(trainer.model.parameters())
			self.optimizer = torch.optim.SGD(self.params, lr=self.kwargs.get("lr", 1.0e-3))
		
		if self.criterion is None and trainer.criterion is not None:
			self.criterion = trainer.criterion
		
		# filter params to get only the ones that require gradients
		self.params = [param for param in self.params if param.requires_grad]
		assert all(param.ndim == 2 for param in self.params), "Eprop only supports 2D parameters for now."
		self.layers_to_params = {
			layer.name: [
				param for param in layer.parameters() if param.requires_grad
			]
			for layer in self.layers
		}
		self.params = [p for params in self.layers_to_params.values() for p in params]
		self.optimizer = torch.optim.SGD(self.params, lr=self.kwargs.get("lr", 1.0e-3))
		
		self.output_layers: torch.nn.ModuleDict = torch.nn.ModuleDict({layer.name: layer for layer in self.layers})
		self._initialize_original_forwards()
		self.initialize_running_grads()
	
	def on_batch_begin(self, trainer, **kwargs):
		super().on_batch_begin(trainer)
		if trainer.model.training:
			if self.feedback_weights is None:
				self.initialize_feedback_weights(self.trainer.current_training_state.y_batch)
			self.eligibility_traces = defaultdict(list)
			self.learning_signals = defaultdict(list)
			self.initialize_running_grads()
			self._last_et = {
				layer.name: [
					torch.zeros_like(p)
					for p in self.layers_to_params[layer.name]
				]
				for layer in self.layers
			}
			# For Debugging
			self.mean_eligibility_traces = defaultdict(list)
			self.mean_learning_signals = defaultdict(list)
			
	def _get_hidden_tensor(self, out: Union[torch.Tensor, Tuple[torch.Tensor], List[torch.Tensor]]):
		if isinstance(out, (tuple, list)):
			out = out[-1]
		return out
	
	def decorate_forwards(self):
		if self.trainer.model.training:
			if not self._forwards_decorated:
				self._initialize_original_forwards()
			for layer in self.output_layers.values():
				layer.forward = self._decorate_forward(layer.forward, layer.name)
			self._forwards_decorated = True
	
	def _decorate_forward(self, forward, layer_name: str):
		def _forward(*args, **kwargs):
			out = forward(*args, **kwargs)
			t = kwargs.get("t", None)
			if t is None:
				return out
			out_tensor = self._get_out_tensor(out)
			h_tensor = self._get_hidden_tensor(out)
			if t == 0:  # Hotfix for the first time step  TODO: fix this
				ready = bool(self._layers_buffer[layer_name])
			else:
				ready = True
			list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
			if len(self._layers_buffer[layer_name]) == self.backward_time_steps and ready:
				self._backward_at_t(t, self.backward_time_steps, layer_name)
			return out
		return _forward
	
	def _backward_at_t(self, t: int, backward_t: int, layer_name: str):
		y_batch = self._get_y_batch_slice_from_trainer((t + 1) - backward_t, t + 1, layer_name)
		# TODO: must have a pred batch for the layer and one for the output_layer to compute the learning signals correctly
		pred_batch = torch.squeeze(self._get_pred_batch_from_buffer(layer_name))
		
		# compute dz/dw
		grad_outputs = torch.eye(pred_batch.shape[-1], device=pred_batch.device)
		params = self.layers_to_params[layer_name]
		instantaneous_eligibility_traces = [torch.zeros_like(p) for p in params]
		# TODO: try to understand what is going on here: https://github.com/IGITUGraz/eligibility_propagation/blob/efd02e6879c01cda3fa9a7838e8e2fd08163c16e/Figure_3_and_S7_e_prop_tutorials/tutorial_pattern_generation.py#L98
		for i in range(grad_outputs.shape[0]):
			zero_grad_params(params)
			
			# pred_batch.backward(grad_outputs[i], retain_graph=True)
			for p_idx, param in enumerate(params):
				# TODO: pas sur du slicing
				instantaneous_eligibility_traces[p_idx][i] = (
					# 0.1 * self._last_et[layer_name][p_idx][i] + param.grad.detach().clone()[i]
					# param.grad.detach().clone()[i]
					torch.autograd.grad(pred_batch[i], param, retain_graph=True)[0][i]
				)
		self._last_et[layer_name] = instantaneous_eligibility_traces
		self.eligibility_traces[layer_name].append(instantaneous_eligibility_traces)
		
		# compute learning signals
		mean_error = torch.mean((y_batch - pred_batch).view(-1, y_batch.shape[-1]), dim=0)
		instantaneous_learning_signals = [
			torch.matmul(mean_error, self.feedback_weights[layer_name][p_idx]).view(1, -1)
			for p_idx in range(len(self.feedback_weights[layer_name]))
		]
		self.learning_signals[layer_name].append(instantaneous_learning_signals)
		self.running_grads[layer_name] = [
			self.running_grads[layer_name][p_idx] + (
					instantaneous_learning_signals[p_idx] * instantaneous_eligibility_traces[p_idx]
			)
			for p_idx in range(len(self.running_grads[layer_name]))
		]
		self._layers_buffer[layer_name].clear()
		
		# For Debugging
		self.mean_eligibility_traces[layer_name].append(
			to_numpy(torch.mean(torch.abs(torch.cat(instantaneous_eligibility_traces, dim=0)))).item()
		)
		self.mean_learning_signals[layer_name].append(
			to_numpy(torch.mean(torch.abs(torch.cat(instantaneous_learning_signals, dim=0)))).item()
		)
	
	def apply_grads(self):
		for layer_name, params in self.layers_to_params.items():
			for p_idx, param in enumerate(params):
				param.grad = self.running_grads[layer_name][p_idx].detach().clone()
	
	def on_optimization_begin(self, trainer, **kwargs):
		y_batch = trainer.current_training_state.y_batch
		pred_batch = trainer.format_pred_batch(trainer.current_training_state.pred_batch, y_batch)
		self.optimizer.zero_grad()
		self.apply_grads()
		self.optimizer.step()
		trainer.update_state_(batch_loss=self.apply_criterion(pred_batch, y_batch).detach_())
	
	def on_optimization_end(self, trainer, **kwargs):
		self.optimizer.zero_grad()
	
	def on_pbar_update(self, trainer, **kwargs) -> dict:
		return {
			# "mean_grads": {
			# 	layer_name: [
			# 		to_numpy(torch.mean(torch.abs(p.grad.detach().clone()))).item()
			# 		for p in self.running_grads[layer_name]
			# 	]
			# 	for layer_name in self.running_grads
			# },
			# "mean_eligibility_traces": self.mean_eligibility_traces,
			# "mean_learning_signals": self.mean_learning_signals,
		}
	



import os
from typing import Iterable, Optional, Tuple, Any

import numpy as np
import torch
from matplotlib import pyplot as plt


@torch.no_grad()
def visualize_init_final_weights(
		initial_wights: Any,
		final_weights: Any,
		filename: Optional[str] = None,
		show: bool = False,
		fig: Optional[plt.Figure] = None,
		axes: Optional[Iterable[plt.Axes]] = None,
		**kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
	import neurotorch as nt
	from mpl_toolkits.axes_grid1.inset_locator import inset_axes
	from matplotlib import colors
	initial_wights, final_weights = nt.to_numpy(initial_wights), nt.to_numpy(final_weights)
	
	assert (fig is None) == (axes is None), "If fig is None, axes must be None"
	if fig is None:
		fig, axes = plt.subplots(1, 2, figsize=(12, 8))
	else:
		assert isinstance(axes, Iterable), "If fig is not None, axes must be an iterable"
	axes_view = np.asarray(axes).ravel()
	
	if kwargs.get("cmp_zero_center", True):
		min_value = min(np.min(initial_wights), np.min(final_weights))
		max_value = max(np.max(initial_wights), np.max(final_weights))
		if not np.isclose(min_value, max_value) and min_value < max_value and min_value < 0 < max_value:
			divnorm = colors.TwoSlopeNorm(
				vmin=min_value,
				vcenter=0.0,
				vmax=max_value
			)
			im = axes_view[0].imshow(initial_wights, cmap="RdBu_r", norm=divnorm)
		else:
			im = axes_view[0].imshow(initial_wights, cmap="RdBu_r")
	else:
		im = axes_view[0].imshow(initial_wights, cmap="RdBu_r")
	axes_view[1].imshow(final_weights, cmap=im.cmap, extent=im.get_extent())
	cax = inset_axes(
		axes_view[1],
		width="5%",
		height="100%",
		loc='lower left',
		bbox_to_anchor=(1.05, 0.0, 1.0, 1.0),
		bbox_transform=axes_view[1].transAxes,
		borderpad=0,
	)
	
	if kwargs.get('compute_dale', "dale_law_kwargs" in kwargs):
		dale_law_kwargs = kwargs.get('dale_law_kwargs', {})
		dale_law_kwargs.setdefault('seed', 0)
		initial_dale = nt.to_numpy(nt.DaleLaw([nt.to_tensor(initial_wights)], **dale_law_kwargs)())
		final_dale = nt.to_numpy(nt.DaleLaw([nt.to_tensor(final_weights)], **dale_law_kwargs)())
		axes_view[0].set_title(f"Initial Weights (Dale loss = {initial_dale:.3f})")
		axes_view[1].set_title(f"Final Weights (Dale loss = {final_dale:.3f})")
	else:
		axes_view[0].set_title("Initial Weights")
		axes_view[1].set_title("Final Weights")
	
	fig.colorbar(im, cax=cax)
	fig.set_tight_layout(False)
	if filename is not None:
		os.makedirs(os.path.dirname(filename), exist_ok=True)
		fig.savefig(filename)
	if show:
		plt.show()
	return fig, axes





from typing import Optional, Tuple

import torch

from neurotorch.dimension import SizeTypes
from neurotorch.modules import BaseLayer


class DimensionsCat:
	def __init__(self, start_axis: int, end_axis: int = -1):
		self.start_axis = start_axis
		self.end_axis = end_axis
		assert self.start_axis < self.end_axis, \
			f"Start axis ({self.start_axis}) must be lower than end axis ({self.end_axis})."
		assert all(axis > 0 for axis in self.axes), \
			"Each axis must be greater than 0. The dimension 0 is the batch dimension and can't be concatenate " \
			"with others."

	@property
	def axes(self):
		return [self.start_axis, self.end_axis]

	def __call__(self, inputs: torch.Tensor):
		assert all(axis < inputs.shape[-1] for axis in self.axes), \
			f"Axes ({self.axes}) are not dimension of inputs ({inputs.shape})."
		return torch.flatten(inputs, start_dim=self.start_axis, end_dim=self.end_axis)


class ReduceSum(BaseLayer):
	def __init__(
			self,
			input_size: Optional[SizeTypes] = None,
			output_size: Optional[SizeTypes] = None,
			name: Optional[str] = None,
			axis: int = -1,
			device: Optional[torch.device] = None,
	):
		super().__init__(
			input_size=input_size,
			output_size=output_size,
			name=name,
			device=device,
		)
		self.axis = axis

	def create_empty_state(self, batch_size: int = 1) -> Tuple[torch.Tensor, ...]:
		pass

	def forward(self, inputs: torch.Tensor, state: torch.Tensor = None) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
		return inputs.sum(dim=self.axis), None


class ReduceMean(BaseLayer):
	def __init__(self, axis: int = -1):
		super().__init__()
		self.axis = axis

	def create_empty_state(self, batch_size: int = 1) -> Tuple[torch.Tensor, ...]:
		pass

	def forward(self, inputs: torch.Tensor, state: torch.Tensor = None) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
		return inputs.mean(dim=self.axis), None











from collections import defaultdict
from typing import Any, Optional, Dict, Callable, Tuple, Union

import numpy as np
import torch
from torch.nn.functional import mse_loss
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from sklearn import metrics as sk_metrics

from .base import BaseMetrics
from ..modules import BaseModel
from ..transforms import to_tensor


class RegressionMetrics(BaseMetrics):
	EPSILON = 1e-12

	@staticmethod
	def get_all_metrics_names_to_func() -> Dict[str, Callable]:
		sep = RegressionMetrics.METRICS_NAMES_SEP
		return {
			f"mae{sep}mean_absolute_error": RegressionMetrics.mean_absolute_error,
			f"mse{sep}mean_squared_error": RegressionMetrics.mean_squared_error,
			f"r2": RegressionMetrics.r2,
			f"d2{sep}d2_tweedie": RegressionMetrics.d2_tweedie,
			f"p_var": RegressionMetrics.p_var,
		}

	@staticmethod
	def compute_y_true_y_pred(
			model: BaseModel,
			dataloader: DataLoader,
			device: Optional[torch.device] = None,
			verbose: bool = False,
			desc: Optional[str] = None,
			p_bar_position: int = 0,
	) -> Union[Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]], Tuple[np.ndarray, np.ndarray]]:
		if device is not None:
			model.to(device)
		model.eval()
		predictions = defaultdict(list)
		targets = defaultdict(list)
		with torch.no_grad():
			for i, (x, y_true) in tqdm(
					enumerate(dataloader), total=len(dataloader),
					desc=desc, disable=not verbose, position=p_bar_position,
			):
				x = x.to(model.device)
				y_true = y_true.to(model.device)
				preds = model.get_prediction_trace(x)
				if isinstance(preds, dict):
					if not isinstance(targets, dict):
						targets = {k: targets for k in preds}
					for k, v in preds.items():
						predictions[k].extend(v.cpu().numpy())
						targets[k].extend(y_true[k].cpu().numpy())
				else:
					predictions["__all__"].extend(preds.cpu().numpy())
					targets["__all__"].extend(y_true.cpu().numpy())
		predictions = {k: np.asarray(v) for k, v in predictions.items()}
		targets = {k: np.asarray(v) for k, v in targets.items()}
		if len(targets) == 1:
			return targets[list(targets.keys())[0]], predictions[list(predictions.keys())[0]]
		return targets, predictions

	@staticmethod
	def mean_absolute_error(
			model: Optional[BaseModel] = None,
			dataloader: Optional[DataLoader] = None,
			y_true: Optional[np.ndarray] = None,
			y_pred: Optional[np.ndarray] = None,
			device: Optional[torch.device] = None,
			verbose: bool = False,
			desc: Optional[str] = None,
			p_bar_position: int = 0,
	) -> Union[float, Dict[str, float]]:
		if y_true is None:
			assert y_pred is None
			assert model is not None, "Either model or y_pred and y_true must be supplied."
			assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
			y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
				model=model, dataloader=dataloader, device=device,
				verbose=verbose, desc=desc, p_bar_position=p_bar_position
			)

		if isinstance(y_true, dict):
			return {k: sk_metrics.mean_absolute_error(y_true[k].flatten(), y_pred[k].flatten()) for k in y_true}
		return sk_metrics.mean_absolute_error(y_true.flatten(), y_pred.flatten())

	@staticmethod
	def mean_squared_error(
			model: Optional[BaseModel] = None,
			dataloader: Optional[DataLoader] = None,
			y_true: Optional[np.ndarray] = None,
			y_pred: Optional[np.ndarray] = None,
			device: Optional[torch.device] = None,
			verbose: bool = False,
			desc: Optional[str] = None,
			p_bar_position: int = 0,
	) -> Union[float, Dict[str, float]]:
		if y_true is None:
			assert y_pred is None
			assert model is not None, "Either model or y_pred and y_true must be supplied."
			assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
			y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
				model, dataloader, device, verbose, desc, p_bar_position
			)

		if isinstance(y_true, dict):
			return {k: sk_metrics.mean_squared_error(y_true[k].flatten(), y_pred[k].flatten()) for k in y_true}
		return sk_metrics.mean_squared_error(y_true.flatten(), y_pred.flatten())

	@staticmethod
	def r2(
			model: Optional[BaseModel] = None,
			dataloader: Optional[DataLoader] = None,
			y_true: Optional[np.ndarray] = None,
			y_pred: Optional[np.ndarray] = None,
			device: Optional[torch.device] = None,
			verbose: bool = False,
			desc: Optional[str] = None,
			p_bar_position: int = 0,
	) -> Union[float, Dict[str, float]]:
		if y_true is None:
			assert y_pred is None
			assert model is not None, "Either model or y_pred and y_true must be supplied."
			assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
			y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
				model, dataloader, device, verbose, desc, p_bar_position
			)

		if isinstance(y_true, dict):
			return {k: sk_metrics.r2_score(y_true[k].flatten(), y_pred[k].flatten()) for k in y_true}
		return sk_metrics.r2_score(y_true.flatten(), y_pred.flatten())

	@staticmethod
	def d2_tweedie(
			model: Optional[BaseModel] = None,
			dataloader: Optional[DataLoader] = None,
			y_true: Optional[np.ndarray] = None,
			y_pred: Optional[np.ndarray] = None,
			device: Optional[torch.device] = None,
			verbose: bool = False,
			desc: Optional[str] = None,
			p_bar_position: int = 0,
	) -> Union[float, Dict[str, float]]:
		if y_true is None:
			assert y_pred is None
			assert model is not None, "Either model or y_pred and y_true must be supplied."
			assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
			y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
				model, dataloader, device, verbose, desc, p_bar_position
			)

		if isinstance(y_true, dict):
			return {k: sk_metrics.d2_tweedie_score(y_true[k].flatten(), y_pred[k].flatten()) for k in y_true}
		return sk_metrics.d2_tweedie_score(y_true.flatten(), y_pred.flatten())

	@staticmethod
	def compute_p_var(
			y_true,
			y_pred,
			device: Optional[torch.device] = None,
			reduction: str = 'mean',
	) -> torch.Tensor:
		y_true, y_pred = to_tensor(y_true), to_tensor(y_pred)
		if device is None:
			device = y_pred.device
		y_true, y_pred = y_true.to(device), y_pred.to(device)
		dims = list(range(y_pred.ndim))[1:]
		mse = torch.mean(mse_loss(y_true, y_pred, reduction='none'), dim=dims)
		var = torch.var(y_true, dim=dims)
		p_var_values = 1 - mse / (var + RegressionMetrics.EPSILON)
		if reduction.lower() == 'none':
			p_var_value = p_var_values
		elif reduction.lower() == 'mean':
			p_var_value = torch.mean(p_var_values)
		elif reduction.lower() == 'sum':
			p_var_value = torch.sum(p_var_values)
		else:
			raise ValueError(f"Reduction {reduction} not recognized. Try 'none'|'sum'|'mean'.")
		return p_var_value

	@staticmethod
	def p_var(
			model: Optional[BaseModel] = None,
			dataloader: Optional[DataLoader] = None,
			y_true: Optional[np.ndarray] = None,
			y_pred: Optional[np.ndarray] = None,
			device: Optional[torch.device] = None,
			verbose: bool = False,
			desc: Optional[str] = None,
			p_bar_position: int = 0,
	) -> Union[float, Dict[str, float]]:
		"""
		Compute the p-value of the variance of the prediction.
		:param model: The model to use for computing the p-value.
		:param dataloader: The dataloader to use for computing the p-value.
		:param y_true: The true values.
		:param y_pred: The predicted values.
		:param device: The device to use for computing the p-value.
		:param verbose: Whether to print progress.
		:param desc: The description to use for the progress bar.
		:param p_bar_position: The position of the progress bar.
		:return: The p-value of the variance of the prediction.
		"""
		if y_true is None:
			assert y_pred is None
			assert model is not None, "Either model or y_pred and y_true must be supplied."
			assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
			y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
				model=model, dataloader=dataloader, device=device,
				verbose=verbose, desc=desc, p_bar_position=p_bar_position
			)

		if isinstance(y_true, dict):
			return {
				k: RegressionMetrics.compute_p_var(y_true[k], y_pred[k], device).detach().cpu().numpy()
				for k in y_true
			}
		return RegressionMetrics.compute_p_var(y_true, y_pred, device).detach().cpu().numpy()

	def __call__(
			self,
			data_loader: DataLoader,
			verbose: Union[bool, int] = False
	) -> Dict[str, Any]:
		"""
		Compute the metrics for the given data_loader.
		:param data_loader: The data loader to use to compute the metrics.
		:param verbose: 0: no progress bar, 1: single progress bar, 2: progress bar for each metrics
						True: verbose is set to 1, False: verbose is set to 0.
		:return: The metrics computed.
		"""
		if isinstance(verbose, bool):
			verbose = 1 if verbose else 0
		assert verbose in [0, 1, 2], "verbose must be 0, 1 or 2"
		output = {}
		p_bar = tqdm(
			enumerate(self.metrics_functions.items()),
			total=len(self.metrics_functions),
			disable=verbose == 0,
			unit="metric",
			position=0,
		)
		self.model.eval()
		y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
			model=self.model,
			dataloader=data_loader,
			device=self.device,
			verbose=verbose == 2,
			desc="compute_y_true_y_preds",
			p_bar_position=1
		)
		for i, (metric_name, metric_func) in p_bar:
			output[metric_name] = metric_func(
				y_true=y_true,
				y_pred=y_pred,
				device=self.device,
				verbose=verbose == 2,
				desc=metric_name,
				p_bar_position=i + 2
			)
		return output



from typing import Optional, Sequence, Union, Dict, Callable, List

import torch

from .learning_algorithm import LearningAlgorithm


class BPTT(LearningAlgorithm):
	r"""
	Apply the backpropagation through time algorithm to the given model.
	"""
	CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"
	
	def __init__(
			self,
			*,
			params: Optional[Sequence[torch.nn.Parameter]] = None,
			layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
			optimizer: Optional[torch.optim.Optimizer] = None,
			criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] = None,
			**kwargs
	):
		"""
		Constructor for BPTT class.
		
		:param params: The parameters to optimize. If None, the parameters of the model's trainer will be used.
		:type params: Optional[Sequence[torch.nn.Parameter]]
		:param optimizer: The optimizer to use. If not provided, torch.optim.Adam is used.
		:type optimizer: Optional[torch.optim.Optimizer]
		:param criterion: The criterion to use. If not provided, torch.nn.MSELoss is used.
		:type criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]]
		:param kwargs: The keyword arguments to pass to the BaseCallback.
		
		:keyword bool save_state: Whether to save the state of the optimizer. Defaults to True.
		:keyword bool load_state: Whether to load the state of the optimizer. Defaults to True.
		"""
		kwargs.setdefault("save_state", True)
		kwargs.setdefault("load_state", True)
		super().__init__(params=params, **kwargs)
		if params is None:
			params = []
		else:
			params = list(params)
		if layers is not None:
			if isinstance(layers, torch.nn.Module):
				layers = [layers]
			params.extend([param for layer in layers for param in layer.parameters() if param not in params])
		self.params: List[torch.nn.Parameter] = params
		self.layers = layers
		self.optimizer = optimizer
		self.criterion = criterion
		
	def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
		if self.save_state:
			state = checkpoint.get(self.name, {})
			opt_state_dict = state.get(self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY, None)
			if opt_state_dict is not None:
				self.optimizer.load_state_dict(opt_state_dict)
			
	def get_checkpoint_state(self, trainer, **kwargs) -> object:
		if self.save_state:
			if self.optimizer is not None:
				return {
					self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: self.optimizer.state_dict()
				}
		return None
	
	def start(self, trainer, **kwargs):
		super().start(trainer)
		if self.params and self.optimizer is None:
			self.optimizer = torch.optim.Adam(self.params)
		elif not self.params and self.optimizer is not None:
			self.params.extend([
				param
				for i in range(len(self.optimizer.param_groups))
				for param in self.optimizer.param_groups[i]["params"]
			])
		else:
			self.params = trainer.model.parameters()
			self.optimizer = torch.optim.Adam(self.params)
		
		if self.criterion is None and trainer.criterion is not None:
			self.criterion = trainer.criterion
	
	def apply_criterion(self, pred_batch, y_batch):
		if self.criterion is None:
			if isinstance(y_batch, dict):
				self.criterion = {key: torch.nn.MSELoss() for key in y_batch}
			else:
				self.criterion = torch.nn.MSELoss()
		
		if isinstance(self.criterion, dict):
			if isinstance(y_batch, torch.Tensor):
				y_batch = {k: y_batch for k in self.criterion}
			if isinstance(pred_batch, torch.Tensor):
				pred_batch = {k: pred_batch for k in self.criterion}
			assert isinstance(pred_batch, dict) and isinstance(y_batch, dict), \
				"If criterion is a dict, pred, y_batch and pred must be a dict too."
			batch_loss = sum([
				self.criterion[k](pred_batch[k], y_batch[k].to(pred_batch[k].device))
				for k in self.criterion
			])
		else:
			if isinstance(pred_batch, dict) and len(pred_batch) == 1:
				pred_batch = pred_batch[list(pred_batch.keys())[0]]
			batch_loss = self.criterion(pred_batch, y_batch.to(pred_batch.device))
		return batch_loss
	
	def _make_optim_step(self, pred_batch, y_batch, retain_graph=False):
		self.optimizer.zero_grad()
		batch_loss = self.apply_criterion(pred_batch, y_batch)
		batch_loss.backward(retain_graph=retain_graph)
		self.optimizer.step()
		return batch_loss.detach_()
	
	def on_optimization_begin(self, trainer, **kwargs):
		y_batch = trainer.current_training_state.y_batch
		pred_batch = trainer.format_pred_batch(trainer.current_training_state.pred_batch, y_batch)
		batch_loss = self._make_optim_step(pred_batch, y_batch)
		trainer.update_state_(batch_loss=batch_loss)
	
	def on_optimization_end(self, trainer, **kwargs):
		self.optimizer.zero_grad()
	
	def on_validation_batch_begin(self, trainer, **kwargs):
		y_batch = trainer.current_training_state.y_batch
		pred_batch = trainer.format_pred_batch(trainer.current_training_state.pred_batch, y_batch)
		batch_loss = self.apply_criterion(pred_batch, y_batch)
		trainer.update_state_(batch_loss=batch_loss)
		
	def extra_repr(self) -> str:
		return f"optimizer={self.optimizer}, criterion={self.criterion}"




from typing import Optional

import numpy as np
import torch


@torch.no_grad()
def dale_(
		tensor: torch.Tensor,
		inh_ratio: float = 0.5,
		rho: float = 0.99,
		inh_first: bool = True,
		seed: Optional[int] = None
) -> torch.Tensor:
	"""
	Initialize a tensor with Dale's law. As default it is half excitatory, half inhibitory. The connections in the
	tensor are considered as i (row) -> j (col).
	
	:param tensor: tensor to initialize. This tensor will be modified inplace.
	:type tensor: torch.Tensor
	:param inh_ratio: ratio of inhibitory connections. Must be between 0 and 1. Default is 0.5.
	:type inh_ratio: float
	:param rho: The connectivity ratio. Must be between 0 and 1. If rho = 1, the tensor will be fully connected.
		Default is 0.99.
	:type rho: float
	:param inh_first: If True, the inhibitory neurons will be in the first half of the tensor. If False, the neurons
		will be shuffled. Default is True.
	:type inh_first: bool
	:param seed: seed for the random number generator. If None, the seed is not set.
	:type seed: Optional[int]
	
	:return: The initialized tensor.
	:rtype: torch.Tensor
	"""
	assert tensor.ndimension() == 2, "tensor must be 2 dimensional."
	assert 0 <= inh_ratio <= 1, "inh_ratio must be between 0 and 1."
	assert 0 <= rho <= 1, "rho must be between 0 and 1."
	rn_gen = torch.Generator()
	if seed is not None:
		rn_gen.manual_seed(seed)
	N, M = tensor.shape
	i, j = torch.triu_indices(N, M)
	N_0 = int((1 - rho) * len(i))  # Number of zero values
	values_upper = torch.cat(
		(torch.zeros(N_0), torch.normal(0, (1 / np.sqrt(N * rho * (1 - rho))), (len(i) - N_0,), generator=rn_gen))
	).to(tensor.device)
	values_lower = torch.cat(
		(torch.zeros(N_0), torch.normal(0, (1 / np.sqrt(N * rho * (1 - rho))), (len(i) - N_0,), generator=rn_gen))
	).to(tensor.device)
	values_upper_rn_indexes = torch.randperm(len(values_upper), generator=rn_gen)
	values_lower_rn_indexes = torch.randperm(len(values_lower), generator=rn_gen)
	tensor[i, j] = values_upper[values_upper_rn_indexes]
	tensor[j, i] = values_lower[values_lower_rn_indexes]
	tensor = torch.abs(tensor)
	if inh_first:
		inh_indexes = torch.arange(int(N * inh_ratio))
	else:
		inh_indexes = torch.randperm(N, generator=rn_gen)[:int(N * inh_ratio)]
	tensor[inh_indexes] *= -1
	return tensor


if __name__ == '__main__':
	from matplotlib import pyplot as plt
	
	plt.imshow(dale_(torch.empty(100, 100), 0.5, 0.99), cmap="RdBu_r")
	plt.colorbar()
	plt.show()





import warnings
from collections import defaultdict
from typing import Optional, Sequence, Union, Dict, Callable, List, Tuple

import torch
from .bptt import BPTT
from ..utils import list_insert_replace_at


class TBPTT(BPTT):
	def __init__(
			self,
			*,
			params: Optional[Sequence[torch.nn.Parameter]] = None,
			layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
			optimizer: Optional[torch.optim.Optimizer] = None,
			criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] = None,
			backward_time_steps: Optional[int] = None,
			optim_time_steps: Optional[int] = None,
			**kwargs
	):
		super(TBPTT, self).__init__(params=params, layers=layers, optimizer=optimizer, criterion=criterion, **kwargs)
		self.output_layers = None
		self._original_forwards = {}
		self._auto_set_backward_time_steps = backward_time_steps is None
		self.backward_time_steps = backward_time_steps
		self._auto_backward_time_steps_ratio = kwargs.get("auto_backward_time_steps_ratio", 0.1)
		assert 0 <= self._auto_backward_time_steps_ratio <= 1, "auto_backward_time_steps_ratio must be between 0 and 1"
		self._auto_set_optim_time_steps = optim_time_steps is None
		self.optim_time_steps = optim_time_steps
		self._auto_optim_time_steps_ratio = kwargs.get("auto_optim_time_steps_ratio", self._auto_backward_time_steps_ratio)
		assert 0 <= self._auto_optim_time_steps_ratio <= 1, "auto_optim_time_steps_ratio must be between 0 and 1"
		self._data_n_time_steps = 0
		self._layers_buffer = defaultdict(list)
		self._forwards_decorated = False
	
	def start(self, trainer, **kwargs):
		super().start(trainer)
		self.output_layers: torch.nn.ModuleDict = trainer.model.output_layers
		self._initialize_original_forwards()
		
	def on_batch_begin(self, trainer, **kwargs):
		super().on_batch_begin(trainer)
		self.trainer = trainer
		if trainer.model.training:
			self._data_n_time_steps = self._get_data_time_steps_from_y_batch(trainer.current_training_state.y_batch)
			self._maybe_update_time_steps()
			self.optimizer.zero_grad()
			self.decorate_forwards()
	
	def on_batch_end(self, trainer, **kwargs):
		super().on_batch_end(trainer)
		if trainer.model.training:
			for layer_name in self._layers_buffer:
				backward_t = len(self._layers_buffer[layer_name])
				if backward_t > 0:
					self._backward_at_t(self._data_n_time_steps - 1, backward_t, layer_name)
					self.optimizer.step()
		self.undecorate_forwards()
		self._layers_buffer.clear()
		self.optimizer.zero_grad()
	
	def _get_data_time_steps_from_y_batch(self, y_batch: Union[torch.Tensor, Dict[str, torch.Tensor]]):
		if isinstance(y_batch, torch.Tensor):
			return y_batch.shape[-2]
		elif isinstance(y_batch, dict):
			return max([y.shape[-2] for y in y_batch.values()])
		else:
			raise ValueError(f"y_batch must be either a torch.Tensor or a dict, but got {type(y_batch)}")
	
	def _initialize_original_forwards(self):
		for layer in self.output_layers.values():
			self._original_forwards[layer.name] = layer.forward
	
	def decorate_forwards(self):
		if self.trainer.model.training:
			if not self._forwards_decorated:
				self._initialize_original_forwards()
			for layer in self.output_layers.values():
				layer.forward = self._decorate_forward(layer.forward, layer.name)
			self._forwards_decorated = True
			
	def undecorate_forwards(self):
		for layer in self.output_layers.values():
			layer.forward = self._original_forwards[layer.name]
		self._forwards_decorated = False
	
	def _maybe_update_time_steps(self):
		if self._auto_set_backward_time_steps:
			self.backward_time_steps = max(1, int(self._auto_backward_time_steps_ratio * self._data_n_time_steps))
		if self._auto_set_optim_time_steps:
			self.optim_time_steps = max(1, int(self._auto_optim_time_steps_ratio * self._data_n_time_steps))
		if self.backward_time_steps != self.optim_time_steps:
			raise NotImplementedError("backward_time_steps != optim_time_steps is not implemented yet")
	
	def _decorate_forward(self, forward, layer_name: str):
		def _forward(*args, **kwargs):
			out = forward(*args, **kwargs)
			t = kwargs.get("t", None)
			if t is None:
				return out
			out_tensor = self._get_out_tensor(out)
			if t == 0:  # Hotfix for the first time step  TODO: fix this
				ready = bool(self._layers_buffer[layer_name])
			else:
				ready = True
			list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
			length = len(self._layers_buffer[layer_name])
			if length == self.backward_time_steps and ready:
				self._backward_at_t(t, self.backward_time_steps, layer_name)
				out = self._detach_out(out)
			if length == self.optim_time_steps and ready:
				self.optimizer.step()
				self.optimizer.zero_grad()
			return out
		return _forward
	
	def _backward_at_t(self, t: int, backward_t: int, layer_name: str):
		y_batch = self._get_y_batch_slice_from_trainer((t + 1) - backward_t, t + 1, layer_name)
		pred_batch = self._get_pred_batch_from_buffer(layer_name)
		batch_loss = self.apply_criterion(pred_batch, y_batch)
		if batch_loss.grad_fn is None:
			raise ValueError(
				f"batch_loss.grad_fn is None. This is probably an internal error. Please report this issue on GitHub."
			)
		batch_loss.backward()
		self._layers_buffer[layer_name].clear()
	
	def _get_y_batch_slice_from_trainer(self, t_first: int, t_last: int, layer_name: str = None):
		y_batch = self.trainer.current_training_state.y_batch
		if isinstance(y_batch, dict):
			if layer_name is None:
				y_batch = {
					key: val[:, t_first:t_last]
					for key, val in y_batch.items()
				}
			else:
				y_batch = y_batch[layer_name][:, t_first:t_last]
		else:
			y_batch = y_batch[:, t_first:t_last]
		return y_batch.clone()
	
	def _get_out_tensor(self, out: Union[torch.Tensor, Tuple[torch.Tensor], List[torch.Tensor]]):
		if isinstance(out, (tuple, list)):
			out = out[0]
		return out
	
	def _detach_out(self, out: Union[torch.Tensor, Tuple[torch.Tensor], List[torch.Tensor]]):
		if isinstance(out, tuple):
			out = tuple([self._detach_out(o) for o in out])
		elif isinstance(out, list):
			out = [self._detach_out(o) for o in out]
		else:
			out = out.detach()
		return out
	
	def _get_pred_batch_from_buffer(self, layer_name: str):
		pred_batch = torch.stack(self._layers_buffer[layer_name], dim=1)
		return pred_batch
	
	def on_optimization_begin(self, trainer, **kwargs):
		y_batch = trainer.current_training_state.y_batch
		pred_batch = trainer.format_pred_batch(trainer.current_training_state.pred_batch, y_batch)
		batch_loss = self.apply_criterion(pred_batch, y_batch)
		trainer.update_state_(batch_loss=batch_loss)
	
	def on_optimization_end(self, trainer, **kwargs):
		super(TBPTT, self).on_optimization_end(trainer)
		self._layers_buffer.clear()
	
	def close(self, trainer, **kwargs):
		self.undecorate_forwards()
		super(TBPTT, self).close(trainer)

